/*
 * Copyright 1996, 1997 Computing Research Labs, New Mexico State University
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE COMPUTING RESEARCH LAB OR NEW MEXICO STATE UNIVERSITY BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
 * OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
#ifndef lint
static char rcsid[] = "$Id: dwidth.c,v 1.1 1997/01/16 04:33:05 mleisher Exp $";
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <X11/IntrinsicP.h>
#include <Xm/Xm.h>
#include <Xm/PushB.h>
#include <Xm/Frame.h>
#include <Xm/Form.h>
#include <Xm/TextF.h>
#include <Xm/DialogS.h>
#include "FGrid.h"
#include "xmbdfed.h"

static void
#ifndef _NO_PROTO
EnableUpdate(Widget w, XtPointer client_data, XtPointer call_data)
#else
EnableUpdate(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];
    XtSetSensitive(ed->dwidth.update, True);
}

static void
#ifndef _NO_PROTO
DoDWidthUpdate(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoDWidthUpdate(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;
    char *s;
    unsigned short dw;

    ed = &editors[(unsigned long) client_data];

    s = XmTextFieldGetString(ed->dwidth.dwidth);
    dw = (unsigned short) _bdf_atoul(s, 0, 10);
    XtFree(s);

    XmuttFontGridSetFontDeviceWidth(ed->fgrid, dw);

    XtSetSensitive(ed->dwidth.update, False);
}

static void
#ifndef _NO_PROTO
DoCloseDWidth(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoCloseDWidth(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];

    if (XtIsSensitive(ed->dwidth.update)) {
        if (AskBooleanQuestion("Device Width has been modified.", "Update",
                               "Do Not Update") == True)
          DoDWidthUpdate(w, client_data, call_data);
    }
    XtPopdown(ed->dwidth.shell);
}

static void
#ifndef _NO_PROTO
BuildDeviceWidth(MXFEditor *ed)
#else
BuildDeviceWidth(ed)
MXFEditor *ed;
#endif
{
    Widget pb, frame, form;
    Cardinal ac;
    Arg av[10];

    if (ed->file == 0)
      sprintf(title, "%s - DevWidth: (unnamed%ld)", app_name, ed->id);
    else
      sprintf(title, "%s - DevWidth: %s", app_name, ed->file);

    ac = 0;
    XtSetArg(av[ac], XmNtitle, title); ac++;
    XtSetArg(av[ac], XmNallowShellResize, True); ac++;
    XtSetArg(av[ac], XmNdeleteResponse, XmUNMAP); ac++;
    sprintf(name, "xmbdfed_dwidth%ld_shell", ed->id);
    ed->dwidth.shell = XtCreatePopupShell(name, xmDialogShellWidgetClass,
                                          top, av, ac);

    ac = 0;
    XtSetArg(av[ac], XmNverticalSpacing, 5); ac++;
    XtSetArg(av[ac], XmNmarginWidth, 2); ac++;
    XtSetArg(av[ac], XmNmarginHeight, 2); ac++;
    sprintf(name, "xmbdfed_dwidth%ld_form", ed->id);
    ed->dwidth.form = form =
        XtCreateWidget(name, xmFormWidgetClass, ed->dwidth.shell, av, ac);

    ac = 0;
    XtSetArg(av[ac], XmNtopAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNleftAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNrightAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNcolumns, 6); ac++;
    sprintf(name, "xmbdfed_dwidth%ld_text", ed->id);
    ed->dwidth.dwidth = XtCreateManagedWidget(name, xmTextFieldWidgetClass,
                                              ed->dwidth.form, av, ac);
    XtAddCallback(ed->dwidth.dwidth, XmNvalueChangedCallback,
                  EnableUpdate, (XtPointer) ed->id);

    ac = 0;
    XtSetArg(av[ac], XmNtopAttachment, XmATTACH_WIDGET); ac++;
    XtSetArg(av[ac], XmNtopWidget, ed->dwidth.dwidth); ac++;
    XtSetArg(av[ac], XmNleftAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNrightAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNbottomAttachment, XmATTACH_FORM); ac++;
    sprintf(name, "xmbdfed_dwidth%ld_frame", ed->id);
    frame = XtCreateManagedWidget(name, xmFrameWidgetClass, ed->dwidth.form,
                                  av, ac);

    ac = 0;
    XtSetArg(av[ac], XmNfractionBase, 5); ac++;
    sprintf(name, "xmbdfed_dwidth%ld_button_form", ed->id);
    form = XtCreateManagedWidget(name, xmFormWidgetClass, frame, av, ac);

    ac = 0;
    XtSetArg(av[ac], XmNtopAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNleftAttachment, XmATTACH_POSITION); ac++;
    XtSetArg(av[ac], XmNleftPosition, 1); ac++;
    XtSetArg(av[ac], XmNrightAttachment, XmATTACH_POSITION); ac++;
    XtSetArg(av[ac], XmNrightPosition, 2); ac++;
    XtSetArg(av[ac], XmNbottomAttachment, XmATTACH_FORM); ac++;
    ed->dwidth.update = XtCreateManagedWidget("Update",
                                              xmPushButtonWidgetClass, form,
                                              av, ac);
    XtAddCallback(ed->dwidth.update, XmNactivateCallback,
                  DoDWidthUpdate, (XtPointer) ed->id);

    ac = 0;
    XtSetArg(av[ac], XmNtopAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNleftAttachment, XmATTACH_POSITION); ac++;
    XtSetArg(av[ac], XmNleftPosition, 3); ac++;
    XtSetArg(av[ac], XmNrightAttachment, XmATTACH_POSITION); ac++;
    XtSetArg(av[ac], XmNrightPosition, 4); ac++;
    XtSetArg(av[ac], XmNbottomAttachment, XmATTACH_FORM); ac++;
    pb = XtCreateManagedWidget("Done", xmPushButtonWidgetClass, form, av, ac);
    XtAddCallback(pb, XmNactivateCallback, DoCloseDWidth, (XtPointer) ed->id);
}

void
#ifndef _NO_PROTO
DoDeviceWidth(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoDeviceWidth(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;
    unsigned short dw;
    Arg av[1];
    char nbuf[25];

    ed = &editors[(unsigned long) client_data];

    if (ed->dwidth.shell == 0)
      BuildDeviceWidth(ed);

    dw = XmuttFontGridFontDeviceWidth(ed->fgrid);
    sprintf(nbuf, "%hd", dw);
    XmTextFieldSetString(ed->dwidth.dwidth, nbuf);

    /*
     * Disable the update button initially.
     */
    XtSetSensitive(ed->dwidth.update, False);

    /*
     * Make sure the title matches the font name.
     */
    if (ed->file == 0)
      sprintf(title, "%s - DevWidth: (unnamed%ld)", app_name, ed->id);
    else
      sprintf(title, "%s - DevWidth: %s", app_name, ed->file);

    XtSetArg(av[0], XmNtitle, title);
    XtSetValues(ed->dwidth.shell, av, 1);

    XtManageChild(ed->dwidth.form);
    XtPopup(ed->dwidth.shell, XtGrabNone);
}

