

/*
 * Copyright 1996, 1997 Computing Research Labs, New Mexico State University
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE COMPUTING RESEARCH LAB OR NEW MEXICO STATE UNIVERSITY BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
 * OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
#ifndef lint
static char rcsid[] = "$Id: xmbdfed.c,v 1.15 1997/10/08 05:30:05 mleisher Exp $";
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <Xm/Xm.h>
#include <Xm/RowColumn.h>
#include <Xm/MainW.h>
#include <Xm/CascadeB.h>
#include <Xm/ToggleB.h>
#include <Xm/PushB.h>
#include <Xm/Label.h>
#include <Xm/Separator.h>
#include <Xm/Frame.h>
#include <Xm/Form.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/MessageB.h>
#include <Xm/FileSB.h>
#include <Xm/DialogS.h>
#include "FGrid.h"
#include "GTest.h"
#include "xmbdfed.h"

/**************************************************************************
 *
 * Global variables.
 *
 **************************************************************************/

Widget top = 0, questd = 0, errd = 0, saved = 0, opend = 0;
XtAppContext app = 0;
Boolean yes_no_answer = False;
Boolean yes_no_done = False;
unsigned long active_editor = 0;
String app_name = 0;
String app_class = 0;
char title[BUFSIZ];
char name[BUFSIZ];

MXFEditor *editors;
unsigned long num_editors = 0;

MXFEditorOptions xmbdfed_opts;

/**************************************************************************
 *
 * Local variables.
 *
 **************************************************************************/

/*
 * The auto-correction message widgets.
 */
static Widget acmsg_shell;
static Widget acmsg_form;
static Widget acmsg_text;

/*
 * Variables for special exit handling.
 */
static XtWorkProcId wpid;
static Boolean save_done;
static Boolean exiting = False;

#define OGINFO "\"%s\" %lo (%o,%o)\nascent %d descent %d right %d left %d"
#define DGINFO "\"%s\" %ld (%d,%d)\nascent %d descent %d right %d left %d"
#define HGINFO "\"%s\" %04lX (%02X,%02X)\nascent %d descent %d right %d left %d"

static String fallback_resources[] = {
    "*fontList: -adobe-new century schoolbook-medium-r-normal--14-100-100-100-p-82-iso8859-1",
    (String) NULL,
};

/*
 * Resources and descriptions for the command line options.
 *
 * IMPORTANT: The order of these resources is important!  Do not
 *            rearrange them unless you know what you are doing!
 */
static XtResource cmdres[] = {
    {
        "noMetricsCorrections", "NoMetricsCorrections",
        XmRInt, sizeof(int),
        XtOffsetOf(MXFEditorOptions, font_opts.correct_metrics),
        XmRImmediate, (XtPointer) 1
    },
    {
        "noUnencodedGlyphs", "NoUnencodedGlyphs",
        XmRInt, sizeof(int),
        XtOffsetOf(MXFEditorOptions, font_opts.keep_unencoded),
        XmRImmediate, (XtPointer) 1
    },
    {
        "noComments", "NoComments",
        XmRInt, sizeof(int),
        XtOffsetOf(MXFEditorOptions, font_opts.keep_comments),
        XmRImmediate, (XtPointer) 1
    },
    {
        "padCharacterCells", "PadCharacterCells",
        XmRInt, sizeof(int),
        XtOffsetOf(MXFEditorOptions, font_opts.pad_cells),
        XmRImmediate, (XtPointer) 1
    },
    {
        "fontPointSize", "FontPointSize",
        XmRInt, sizeof(long),
        XtOffsetOf(MXFEditorOptions, font_opts.point_size),
        XmRImmediate, (XtPointer) 0
    },
    {
        "fontHorizontalResolution", "FontHorizontalResolution",
        XmRInt, sizeof(unsigned long),
        XtOffsetOf(MXFEditorOptions, font_opts.resolution_x),
        XmRImmediate, (XtPointer) 0
    },
    {
        "fontVerticalResolution", "FontVerticalResolution",
        XmRInt, sizeof(unsigned long),
        XtOffsetOf(MXFEditorOptions, font_opts.resolution_y),
        XmRImmediate, (XtPointer) 0
    },
    {
        "fontSpacing", "FontSpacing",
        "FontSpacing", sizeof(int),
        XtOffsetOf(MXFEditorOptions, font_opts.font_spacing),
        XmRImmediate, (XtPointer) 0
    },
    {
        "fontResolution", "FontResolution",
        XmRInt, sizeof(unsigned long),
        XtOffsetOf(MXFEditorOptions, resolution),
        XmRImmediate, (XtPointer) 0
    },
    {
        "skipBlankGlyphPages", "SkipBlankGlyphPages",
        XmRInt, sizeof(int),
        XtOffsetOf(MXFEditorOptions, no_blanks),
        XmRImmediate, (XtPointer) 1
    },
    {
        "reallyExitDialog", "ReallyExitDialog",
        XmRInt, sizeof(int),
        XtOffsetOf(MXFEditorOptions, really_exit),
        XmRImmediate, (XtPointer) 1
    },
    {
        "gridOverwriteMode", "gridOverwriteMode",
        XmRInt, sizeof(int),
        XtOffsetOf(MXFEditorOptions, overwrite_mode),
        XmRImmediate, (XtPointer) 1
    },
};

static XrmOptionDescRec cmdopts[] = {
    {"-nm",   "*noMetricsCorrections", XrmoptionNoArg, (XtPointer) "0"},
    {"-nu",   "*noUnencodedGlyphs", XrmoptionNoArg, (XtPointer) "0"},
    {"-nc",   "*noComments", XrmoptionNoArg, (XtPointer) "0"},
    {"-np",   "*padCharacterCells", XrmoptionNoArg, (XtPointer) "0"},
    {"-bp",   "*skipBlankGlyphPages", XrmoptionNoArg, (XtPointer) "0"},
    {"-ps",   "*fontPointSize", XrmoptionSepArg, (XtPointer) "0"},
    {"-hres", "*fontHorizontalResolution", XrmoptionSepArg, (XtPointer) "0"},
    {"-vres", "*fontVerticalResolution", XrmoptionSepArg, (XtPointer) "0"},
    {"-res",  "*fontResolution", XrmoptionSepArg, (XtPointer) "0"},
    {"-sp",   "*fontSpacing", XrmoptionSepArg, (XtPointer) "0"},
    {"-ed",   "*reallyExitDialog", XrmoptionNoArg, (XtPointer) "0"},
    {"-im",   "*gridOverwriteMode", XrmoptionNoArg, (XtPointer) "0"},
};

/**************************************************************************
 *
 * Prototypes.
 *
 **************************************************************************/

static unsigned long MakeEditor(
#ifndef _NO_PROTO
    Widget parent,
    char *filename,
    Boolean cmdline
#endif
);

/**************************************************************************
 *
 * Functions.
 *
 **************************************************************************/

void
#ifndef _NO_PROTO
DoClose(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoClose(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    XtPopdown((Widget) client_data);
}

void
#ifndef _NO_PROTO
ErrorDialog(char *message)
#else
ErrorDialog(message)
char *message;
#endif
{
    XmString s;
    Arg av[1];

    if (errd == 0) {
        sprintf(title, "%s: Error Dialog", app_name);
        XtSetArg(av[0], XmNtitle, title);
        errd = XmCreateErrorDialog(top, "ucedit_error_dialog", av, 1);
        XtUnmanageChild(XmMessageBoxGetChild(errd, XmDIALOG_HELP_BUTTON));
        XtUnmanageChild(XmMessageBoxGetChild(errd, XmDIALOG_CANCEL_BUTTON));
    }
    s = XmStringCreateLtoR(message, XmSTRING_DEFAULT_CHARSET);
    XtVaSetValues(errd, XmNmessageString, s, 0);
    XmStringFree(s);
    XtManageChild(errd);
    XtPopup(XtParent(errd), XtGrabNone);
}

static void
#ifndef _NO_PROTO
SetBooleanAnswer(Widget w, XtPointer client_data, XtPointer call_data)
#else
SetBooleanAnswer(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    unsigned long n;

    n = (unsigned long) client_data;

    yes_no_answer = (n) ? True : False;
    yes_no_done = True;
}

Boolean
#ifndef _NO_PROTO
AskBooleanQuestion(char *question, char *ok_label, char *cancel_label)
#else
AskBooleanQuestion(question, ok_label, cancel_label)
char *question, *ok_label, *cancel_label;
#endif
{
    XmString s, cl, ol;
    Cardinal ac;
    Arg av[3];

    if (questd == 0) {
        sprintf(title, "%s: Question", app_name);
        XtSetArg(av[0], XmNtitle, title);
        questd = XmCreateQuestionDialog(top, "ucedit_question_dialog",
                                        av, 1);
        XtUnmanageChild(XmMessageBoxGetChild(questd, XmDIALOG_HELP_BUTTON));
        XtAddCallback(questd, XmNokCallback, SetBooleanAnswer, (XtPointer) 1);
        XtAddCallback(questd, XmNcancelCallback, SetBooleanAnswer,
                      (XtPointer) 0);
    }
    cl = ol = 0;
    ac = 0;
    s = XmStringCreateLtoR(question, XmSTRING_DEFAULT_CHARSET);
    XtSetArg(av[ac], XmNmessageString, s); ac++;
    if (ok_label != 0) {
        ol = XmStringCreateSimple(ok_label);
        XtSetArg(av[ac], XmNokLabelString, ol); ac++;
    }
    if (cancel_label != 0) {
        cl = XmStringCreateSimple(cancel_label);
        XtSetArg(av[ac], XmNcancelLabelString, cl); ac++;
    }
    XtSetValues(questd, av, ac);
    XmStringFree(s);
    if (ol != 0)
        XmStringFree(ol);
    if (cl != 0)
        XmStringFree(cl);

    XtManageChild(questd);
    XtPopup(XtParent(questd), XtGrabNone);

    yes_no_done = False;
    while (yes_no_done == False)
      XtAppProcessEvent(app, XtIMAll);

    XtUnmanageChild(questd);
    XtPopdown(XtParent(questd));
    return yes_no_answer;
}

static void
#ifndef _NO_PROTO
ReallySaveAs(Widget w, XtPointer client_data, XtPointer call_data)
#else
ReallySaveAs(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    FILE *out;
    char *path, *file, *fp;
    MXFEditor *ed;
    XmFileSelectionBoxCallbackStruct *cb;
    bdf_property_t vanity;

    ed = &editors[active_editor];

    path = 0;

    save_done = True;

    cb = (XmFileSelectionBoxCallbackStruct *) call_data;
    XmStringGetLtoR(cb->value, XmSTRING_DEFAULT_CHARSET, &path);
    out = fopen(path, "w");
    setbuf(out, 0);
    if (out == 0) {
        sprintf(name, "Unable to save to '%s'.", path);
        ErrorDialog(name);
        exiting = False;
    } else {
        /*
         * Add the xmbdfed info property before saving the font, but only if
         * the font was really modified by the editor.
         */
        if (XmuttFontGridModified(ed->fgrid) == True) {
            sprintf(name, "Edited with xmbdfed %s.", XMBDFED_VERSION);
            vanity.name = "_XMBDFED_INFO";
            vanity.format = BDF_ATOM;
            vanity.value.atom = name;
            bdf_add_font_property(XmuttFontGridFont(ed->fgrid), &vanity);
        }
        if ((fp = strrchr(path, '/')) == 0)
          fp = path;
        else
          fp++;
        bdf_save_font(out, XmuttFontGridFont(ed->fgrid),
                      &xmbdfed_opts.font_opts, UpdateProgressBar, (void *) fp);
        fclose(out);
    }

    /*
     * Clear the old directory and filename and
     * update with the new ones.
     */
    if (ed->path != 0)
      XtFree(ed->path);
    if (ed->file != 0)
      XtFree(ed->file);

    file = strrchr(path, '/');
    if (file) {
        *file++ = 0;
        ed->path = XtMalloc(strlen(path) + 1);
        (void) strcpy(ed->path, path);
    } else {
        file = path;
        ed->path = XtMalloc(2);
        ed->path[0] = '.';
        ed->path[1] = 0;
    }
    ed->file = XtMalloc(strlen(file) + 1);
    (void) strcpy(ed->file, file);

    /*
     * Update the title of the editor window.
     */
    sprintf(title, "%s - %s", app_name, file);
    XtVaSetValues(ed->shell, XmNtitle, title, 0);

    /*
     * Free the path give back by the file selection dialog.
     */
    XtFree(path);

    /*
     * Desensitize the save button on the menu.
     */
    XtSetSensitive(ed->filemenu.save, False);

    /*
     * Tell the font grid the font has been saved.
     */
    XmuttFontGridSetModified(ed->fgrid, False);

    XtPopdown(XtParent(saved));
}

static void
#ifndef _NO_PROTO
DoCancelSave(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoCancelSave(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    save_done = True;
    XtPopdown((Widget) client_data);
}

static void
#ifndef _NO_PROTO
DoSaveAs(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoSaveAs(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    int pos1, pos2;
    unsigned long id;
    MXFEditor *ed;
    Widget fstext;
    char *dir, *dp;
    XmString s;
    Boolean free_dir;
    Arg av[2];

    id = (unsigned long) client_data;
    ed = &editors[id];

    if (saved == 0) {
        sprintf(title, "%s: Save File", app_name);
        s = XmStringCreateSimple("*.[Bb][Dd][Ff]");
        XtSetArg(av[0], XmNtitle, title);
        XtSetArg(av[1], XmNpattern, s);
        saved = XmCreateFileSelectionDialog(top, "xmbdfed_save_file_sel_box",
                                            av, 2);
        XmStringFree(s);
        XtAddCallback(saved, XmNokCallback, ReallySaveAs, 0);
        XtAddCallback(saved, XmNcancelCallback, DoCancelSave,
                      (XtPointer) XtParent(saved));
        XtManageChild(saved);
    }

    active_editor = id;

    free_dir = False;
    dir = 0;
    if (ed->file != 0) {
        dir = ed->path;
        sprintf(name, "%s", ed->file);
    } else
      sprintf(name, "unnamed%ld.bdf", id);

    if (dir == 0) {
        fstext = XmFileSelectionBoxGetChild(saved, XmDIALOG_FILTER_TEXT);
        dir = XmTextFieldGetString(fstext);
        if ((dp = strrchr(dir, '/')))
          *dp = 0;
        free_dir = True;
    }

    /*
     * Set the filename in the file selection box.
     */
    fstext = XmFileSelectionBoxGetChild(saved, XmDIALOG_TEXT);
    sprintf(title, "%s/%s", dir, name);
    XmTextFieldSetString(fstext, title);
    pos1 = strlen(dir) + 1;
    pos2 = pos1 + strlen(name);
    XmTextFieldSetSelection(fstext, pos1, pos2, CurrentTime);

    if (free_dir)
      XtFree(dir);

    XtPopup(XtParent(saved), XtGrabNone);
}

static void
#ifndef _NO_PROTO
DoSave(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoSave(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    FILE *out;
    MXFEditor *ed;
    unsigned long id;
    bdf_property_t vanity;
    char path[BUFSIZ];

    id = (unsigned long) client_data;
    ed = &editors[id];

    if (XmuttFontGridModified(ed->fgrid) == False)
      return;

    if (ed->file == 0) {
        DoSaveAs(w, client_data, call_data);
        return;
    }

    save_done = True;

    if (ed->path != 0)
      sprintf(path, "%s/%s", ed->path, ed->file);
    else
      sprintf(path, "%s", ed->file);

    out = fopen(path, "w");
    if (out == 0) {
        sprintf(name, "Unable to save to '%s'.", path);
        ErrorDialog(name);
        exiting = False;
        return;
    }

    /*
     * Add the xmbdfed info property before saving the font, but only if
     * the font was really modified by the editor.
     */
    if (XmuttFontGridModified(ed->fgrid) == True) {
        sprintf(name, "Edited with xmbdfed %s.", XMBDFED_VERSION);
        vanity.name = "_XMBDFED_INFO";
        vanity.format = BDF_ATOM;
        vanity.value.atom = name;
        bdf_add_font_property(XmuttFontGridFont(ed->fgrid), &vanity);
    }
    bdf_save_font(out, XmuttFontGridFont(ed->fgrid), &xmbdfed_opts.font_opts,
                  UpdateProgressBar, (void *) ed->file);

    fclose(out);

    /*
     * Update the title of the editor window.
     */
    sprintf(title, "%s - %s", app_name, ed->file);
    XtVaSetValues(ed->shell, XmNtitle, title, 0);

    /*
     * Desensitize the save button on the menu.
     */
    XtSetSensitive(ed->filemenu.save, False);

    /*
     * Tell the font grid the font has been saved.
     */
    XmuttFontGridSetModified(ed->fgrid, False);
}

void
#ifndef _NO_PROTO
WaitSave(Widget w, MXFEditor *ed)
#else
WaitSave(w, ed)
Widget w;
MXFEditor *ed;
#endif
{
    if (ed->file == 0)
      sprintf(name,
              "\"unnamed%ld\" has been modified.\nDo you want to save?",
              ed->id);
    else
        sprintf(name, "\"%s\" has been modified.\nDo you want to save?",
                ed->file);

    if (AskBooleanQuestion(name, "Yes", "No") == True) {
        save_done = False;
        DoSave(w, (XtPointer) ed->id, 0);
        while (save_done == False)
          XtAppProcessEvent(app, XtIMAll);
    }
}

static void
#ifndef _NO_PROTO
DoCloseEditor(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoCloseEditor(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];

    /*
     * Close all open glyph editors associated with this editor.
     */
    GlyphEditClose(ed);

    /*
     * Now check to see if the font should be saved or not.
     */
    if (XmuttFontGridModified(ed->fgrid) == True)
      WaitSave(w, ed);
            
    XtPopdown(ed->shell);
}

static void
#ifndef _NO_PROTO
ShowEditor(Widget w, XtPointer client_data, XtPointer call_data)
#else
ShowEditor(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    unsigned long id;

    id = (unsigned long) client_data;

    if (id > 0)
      XtPopup(editors[id].shell, XtGrabNone);

    /*
     * Force the font grid to have the focus.
     */
    (void) XmProcessTraversal(editors[id].fgrid, XmTRAVERSE_CURRENT);

    XMapRaised(XtDisplay(editors[id].shell), XtWindow(editors[id].shell));
}

/*
 * Determine if this file is an HBF font or not.
 */
static int
#ifndef _NO_PROTO
IsHBFFont(char *filename, int gui_active)
#else
IsHBFFont(filename, gui_active)
char *filename;
int gui_active;
#endif
{
    int hbf;
    FILE *in;
    char *ext, hdr[16];

    if ((ext = strrchr(filename, '.')) == 0) {
        /*
         * There is no extension on the file, so open it and read the first 14
         * characters to determine if it is an HBF or not.
         */
        if ((in = fopen(filename, "r")) == 0) {
            if (gui_active != 0) {
                sprintf(name, "Unable to open font '%s'.", filename);
                ErrorDialog(name);
            } else
              fprintf(stderr, "%s: unable to open font '%s'.\n",
                      app_name, filename);
            return -1;
        }
        fread(hdr, 14, 1, in);
        fclose(in);

        /*
         * If the header string does not belong to either an HBF or a BDF
         * font, return an error code.
         */
        hbf = (memcmp(hdr, "HBF_START_FONT", 14) == 0) ? 1 : 0;
        if (hbf == 0 && memcmp(hdr, "STARTFONT", 9) != 0) {
            if (gui_active != 0) {
                sprintf(name, "Not a BDF or HBF font '%s'.", filename);
                ErrorDialog(name);
            } else
              fprintf(stderr, "%s: Not a BDF or HBF font '%s'.\n",
                      app_name, filename);
            return -1;
        }
        return hbf;
    }

    /*
     * There is an extension, so check to see if it is an HBF extension.
     */
    ext++;
    return ((*ext == 'h' || *ext == 'H') &&
            (*(ext + 1) == 'b' || *(ext + 1) == 'B') &&
            (*(ext + 2) == 'f' || *(ext + 2) == 'F')) ? 1 : 0;
}

static int
#ifndef _NO_PROTO
LoadHBF(char **path, MXFEditor *ed, bdf_font_t **font)
#else
LoadHBF(path, ed, font)
char **path;
MXFEditor *ed;
bdf_font_t **font;
#endif
{
    int plen;
    char *pp, *pe, *file, *ext;
    bdf_font_t *fp;

    /*
     * Get the file name part of the path.
     */
    pp = *path;
    plen = strlen(pp);
    if ((file = strrchr(pp, '/')) == 0)
      file = pp;
    else
      file++;

    if ((fp = bdf_load_hbf_font(pp, &xmbdfed_opts.font_opts,
                                UpdateProgressBar, (void *) file)) == 0) {
        sprintf(name, "Problem loading HBF font '%s'.", file);
        ErrorDialog(name);
        return 0;
    }

    /*
     * Set up to replace an existing extension with ".bdf" or add a ".bdf"
     * extension.
     */
    if ((ext = strrchr(file, '.')) != 0)
      ext++;
    else
      ext = pp + plen;

    /*
     * Make sure the path has enough storage to hold a ".bdf" extension.
     */
    pe = pp + plen;
    if (pe - ext < 3)
      *path = pp = XtRealloc(pp, plen + 4);

    /*
     * Add or replace the extension.
     */
    if ((ext = strrchr(pp, '.')) == 0)
      ext = pp + plen;
    sprintf(ext, ".bdf");

    *font = fp;
    return 1;
}

/*
 * Load a Linux console or Sun VF font.  There can be up to 3 fonts loaded by
 * one call.
 */
static int
#ifndef _NO_PROTO
LoadConsoleFont(char **path, MXFEditor *ed)
#else
LoadConsoleFont(path, ed)
char **path;
MXFEditor *ed;
#endif
{
    FILE *in;
    MXFEditor *ep;
    unsigned long id, eds[2];
    int i, nfonts, plen, len;
    char *pp, *file, *ext, *npath;
    bdf_font_t *fonts[3];
    Arg av[2];
    XmuttFontGridPageInfoStruct pi;

    /*
     * Save the ID of the editor passed so it can be used later.
     */
    id = ed->id;

    /*
     * Get the file name part of the path.
     */
    pp = *path;
    plen = strlen(pp);
    if ((file = strrchr(pp, '/')) == 0)
      file = pp;
    else
      file++;

    if ((in = fopen(pp, "r")) == 0) {
        sprintf(name, "Unable to open console font '%s'.", file);
        ErrorDialog(name);
        return 0;
    }

    if (bdf_load_console_font(in, &xmbdfed_opts.font_opts,
                              UpdateProgressBar, (void *) file,
                              fonts, &nfonts)) {
        fclose(in);
        sprintf(name, "Problem loading console font '%s'.", file);
        ErrorDialog(name);
        return 0;
    }
    fclose(in);

    /*
     * Munge the path so we can generate nice file and font names with the
     * point sizes in them.
     */
    npath = (char *) XtMalloc(plen + 8);
    (void) memcpy(npath, pp, plen + 1);

    if ((file = strrchr(npath, '/')) == 0)
      file = npath;
    else
      *file++ = 0;

    if ((ext = strrchr(file, '.')) == 0)
      ext = npath + plen;
    *ext = 0;

    for (i = 0; i < nfonts; i++) {
        if (i) {
            /*
             * Create the editor for the next smaller size font.
             */
            eds[i - 1] = MakeEditor(top, 0, False);
            ep = &editors[eds[i - 1]];

            /*
             * Creating new editors may move the editor list around in memory,
             * so reset the pointer to the original editor passed.
             */
            ed = &editors[id];
        } else
          /*
           * Point at the initial editor.
           */
          ep = ed;

        /*
         * Create the default font name.
         */
        if (nfonts == 3) {
            switch (i) {
              case 0: sprintf(ext, "-16"); break;
              case 1: sprintf(ext, "-14"); break;
              case 2: sprintf(ext, "-08"); break;
            }
        }

        /*
         * Set the default font name.
         */
        len = strlen(file);
        fonts[i]->name = (char *) malloc(len + 1);
        (void) memcpy(fonts[i]->name, file, len + 1);

        /*
         * Set the font directory and filename.
         */
        if (nfonts == 3) {
            switch (i) {
              case 0: sprintf(ext, "-16.bdf"); break;
              case 1: sprintf(ext, "-14.bdf"); break;
              case 2: sprintf(ext, "-08.bdf"); break;
            }
        } else
          sprintf(ext, ".bdf");

        /*
         * Configure the editor.
         */
        XtSetArg(av[0], XmNbdfFont, fonts[i]);
        XtSetArg(av[1], XmNdisplayUnencoded, False);
        XtSetValues(ep->fgrid, av, 2);

        ep->update_fontname = False;
        XmTextFieldSetString(ep->fontname, XmuttFontGridFontName(ep->fgrid));
        ep->update_fontname = True;

        /*
         * Enable or disable the navigation buttons based on the page
         * info of the font.
         */
        XmuttFontGridPageInfo(ep->fgrid, &pi);

        if (pi.previous_page < 0) {
            XtSetSensitive(ep->prevpage, False);
            XtSetSensitive(ep->firstpage, False);
        } else {
            XtSetSensitive(ep->prevpage, True);
            XtSetSensitive(ep->firstpage, True);
        }

        if (pi.next_page < 0) {
            XtSetSensitive(ep->nextpage, False);
            XtSetSensitive(ep->lastpage, False);
        } else {
            XtSetSensitive(ep->nextpage, True);
            XtSetSensitive(ep->lastpage, True);
        }

        /*
         * Set the current page number in the text field.
         */
        sprintf(name, "%ld", pi.current_page);
        XmTextFieldSetString(ep->pageno, name);
        XmTextFieldSetCursorPosition(ep->pageno, strlen(name));

        /*
         * Enable or disable the Messages button.
         */
        if (XmuttFontGridFontMessages(ep->fgrid) == 0)
          XtSetSensitive(ep->viewmenu.acmsgs, False);
        else
          XtSetSensitive(ep->viewmenu.acmsgs, True);

        /*
         * Change the grid to indicate it has been modified.
         */
        XmuttFontGridSetModified(ep->fgrid, True);
    }

    /*
     * Get rid of the path copy and reset the location of the file pointer
     * in the original path so it can be used later.
     */
    XtFree(npath);

    /*
     * Make sure the open dialog is popped down.
     */
    XtPopdown(XtParent(opend));

    return 1;
}

/*
 * Load a Metafont PK or GF font.
 */
static int
#ifndef _NO_PROTO
LoadPKGFFont(char **path, MXFEditor *ed, bdf_font_t **font)
#else
LoadPKGFFont(path, ed, font)
char **path;
MXFEditor *ed;
bdf_font_t **font;
#endif
{
    FILE *in;
    int plen;
    char *pp, *file, *ext, *pe;
    bdf_font_t *fp;

    /*
     * Get the file name part of the path.
     */
    pp = *path;
    plen = strlen(pp);
    if ((file = strrchr(pp, '/')) == 0)
      file = pp;
    else
      file++;

    if ((in = fopen(pp, "r")) == 0) {
        sprintf(name, "Unable to open PK/GF font '%s'.", file);
        ErrorDialog(name);
        return 0;
    }

    if (bdf_load_mf_font(in, &xmbdfed_opts.font_opts,
                         UpdateProgressBar, (void *) file, &fp)) {
        fclose(in);
        sprintf(name, "Problem loading PK/GF font '%s'.", file);
        ErrorDialog(name);
        return 0;
    }
    fclose(in);

    *font = fp;

    /*
     * Need to munge the font name and file name around to get something
     * usable.
     */
    if ((ext = strrchr(file, '.')) != 0)
      ext++;
    else
      ext = pp + plen;

    /*
     * Make sure the path has enough storage to hold the new file name.
     */
    pe = pp + plen;
    if (pe - ext < 9)
      *path = pp = XtRealloc(pp, plen + 10);

    /*
     * Locate the file name again.
     */
    if ((file = strrchr(pp, '/')) == 0)
      file = pp;
    else
      file++;

    /*
     * Add or replace the rest of the file name.
     */
    if ((ext = strrchr(file, '.')) == 0)
      ext = pp + plen;

    /*
     * Use the horizontal resolution of the font as part of the name to
     * distinguish those with the same name.
     */
    sprintf(ext, "-%ld.bdf", fp->resolution_x);

    /*
     * Locate the extension again.
     */
    if ((ext = strrchr(file, '.')) == 0) {
        /*
         * Use the filename as the font name.
         */
        plen = strlen(file) + 1;
        fp->name = malloc(plen);
        (void) memcpy(fp->name, file, plen);
    } else {
        /*
         * Copy everything up to the extension.
         */
        fp->name = malloc((ext - file) + 1);
        (void) memcpy(fp->name, file, ext - file);
        fp->name[ext - file] = 0;
    }

    /*
     * Make sure the open dialog is popped down.
     */
    XtPopdown(XtParent(opend));

    return 1;
}

static void
#ifndef _NO_PROTO
ReallyOpen(Widget w, XtPointer client_data, XtPointer call_data)
#else
ReallyOpen(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;
    char *path, *file, *fp;
    FILE *in;
    bdf_font_t *f;
    XmFileSelectionBoxCallbackStruct *cb;
    XmString s;
    XmuttFontGridPageInfoStruct pi;
    Arg av[2];

    ed = &editors[active_editor];
    cb = (XmFileSelectionBoxCallbackStruct *) call_data;

    path = 0;
    XmStringGetLtoR(cb->value, XmSTRING_DEFAULT_CHARSET, &path);

    switch (ed->filemenu.open_type) {
      case XMBDFED_OPEN_BDF:
        /*
         * Set the filename that will be used for the progress bar.
         */
        if ((fp = strrchr(path, '/')) == 0)
          fp = path;
        else
          fp++;

        if ((in = fopen(path, "r")) == 0) {
            sprintf(name, "Unable to open BDF font '%s'.", path);
            ErrorDialog(name);
            XtFree(path);
            return;
        }
        f = bdf_load_font(in, &xmbdfed_opts.font_opts, UpdateProgressBar,
                          (void *) fp);
        fclose(in);
        if (f == 0) {
            sprintf(name, "Problem loading BDF font '%s'.", path);
            ErrorDialog(name);
            XtFree(path);
            return;
        }
        break;
      case XMBDFED_OPEN_HBF:
        if (!LoadHBF(&path, ed, &f)) {
            XtFree(path);
            return;
        }
        break;
      case XMBDFED_OPEN_CONSOLE:
        /*
         * This routine takes care of updating all the font grids because
         * there may be 1 or 3 fonts.
         */
        (void) LoadConsoleFont(&path, ed);
        XtFree(path);
        return;
      case XMBDFED_OPEN_PKGF:
        if (!LoadPKGFFont(&path, ed, &f)) {
            XtFree(path);
            return;
        }
        break;
    }

    /*
     * If the glyph test dialog uses the font that is about to be
     * replaced, erase it.
     */
    GlyphTestReset(XmuttFontGridFont(ed->fgrid));

    /*
     * Determine the new path and filename.
     */
    if (ed->path != 0)
      XtFree(ed->path);
    if (ed->file != 0)
      XtFree(ed->file);

    file = strrchr(path, '/');
    if (file) {
        *file++ = 0;
        ed->path = XtMalloc(strlen(path) + 1);
        (void) strcpy(ed->path, path);
    } else {
        file = path;
        ed->path = XtMalloc(2);
        ed->path[0] = '.';
        ed->path[1] = 0;
    }
    ed->file = XtMalloc(strlen(file) + 1);
    (void) strcpy(ed->file, file);

    /*
     * Free up the path from the file selection dialog.
     */
    XtFree(path);

    /*
     * Change the font in the font editor.
     */
    XtSetArg(av[0], XmNbdfFont, f);
    XtSetArg(av[1], XmNdisplayUnencoded, False);
    XtSetValues(ed->fgrid, av, 2);

    /*
     * Update the font name and glyph labels.
     */
    ed->update_fontname = False;
    XmTextFieldSetString(ed->fontname, XmuttFontGridFontName(ed->fgrid));
    ed->update_fontname = True;

    s = XmStringCreateLtoR("None\nNone", XmSTRING_DEFAULT_CHARSET);
    XtSetArg(av[0], XmNlabelString, s);
    XtSetValues(ed->glyphinfo, av, 1);
    XmStringFree(s);

    /*
     * Enable or disable the navigation buttons based on the page info
     * of the font.
     */
    XmuttFontGridPageInfo(ed->fgrid, &pi);

    if (pi.previous_page < 0) {
        XtSetSensitive(ed->prevpage, False);
        XtSetSensitive(ed->firstpage, False);
    } else {
        XtSetSensitive(ed->prevpage, True);
        XtSetSensitive(ed->firstpage, True);
    }

    if (pi.next_page < 0) {
        XtSetSensitive(ed->nextpage, False);
        XtSetSensitive(ed->lastpage, False);
    } else {
        XtSetSensitive(ed->nextpage, True);
        XtSetSensitive(ed->lastpage, True);
    }

    /*
     * Make sure the View menu is updated so the Encoded/Unencoded
     * menu item says the correct thing.
     */
    s = XmStringCreateSimple("Unencoded");
    XtSetArg(av[0], XmNlabelString, s);
    XtSetValues(ed->viewmenu.unencoded, av, 1);
    XmStringFree(s);

    /*
     * Set the current page number in the text field.
     */
    sprintf(name, "%ld", pi.current_page);
    XmTextFieldSetString(ed->pageno, name);
    XmTextFieldSetCursorPosition(ed->pageno, strlen(name));

    /*
     * Enable or disable the Messages button.
     */
    if (XmuttFontGridFontMessages(ed->fgrid) == 0)
      XtSetSensitive(ed->viewmenu.acmsgs, False);
    else
      XtSetSensitive(ed->viewmenu.acmsgs, True);

    /*
     * Enable the Save menu option if the font has been modified.  Also update
     * the title.
     */
    if (XmuttFontGridModified(ed->fgrid) == True) {
        if (ed->file == 0)
          sprintf(title, "%s - (unnamed%ld) [modified]", app_name, ed->id);
        else
          sprintf(title, "%s - %s [modified]", app_name, ed->file);
        XtSetArg(av[0], XmNtitle, title);
        XtSetValues(ed->shell, av, 1);
        XtSetSensitive(ed->filemenu.save, True);
    } else {
        if (ed->file == 0)
          sprintf(title, "%s - (unnamed%ld)", app_name, ed->id);
        else
          sprintf(title, "%s - %s", app_name, ed->file);
        XtSetArg(av[0], XmNtitle, title);
        XtSetValues(ed->shell, av, 1);
        XtSetSensitive(ed->filemenu.save, False);
    }

    /*
     * Hide the file selection dialog.
     */
    XtPopdown(XtParent(opend));
}

static void
#ifndef _NO_PROTO
OpenImportDialog(Widget w, MXFEditor *ed)
#else
OpenImportDialog(w, ed)
Widget w;
MXFEditor *ed;
#endif
{
    XmString s;
    Arg av[2];

    if (XmuttFontGridModified(ed->fgrid) == True)
      WaitSave(w, ed);

    if (opend == 0) {
        switch (ed->filemenu.open_type) {
          case XMBDFED_OPEN_BDF:
            sprintf(title, "%s: Open BDF Font", app_name);
            s = XmStringCreateSimple("*.[Bb][Dd][Ff]");
            break;
          case XMBDFED_OPEN_HBF:
            sprintf(title, "%s: Import HBF Font", app_name);
            s = XmStringCreateSimple("*.[Hh][Bb][Ff]");
            break;
          case XMBDFED_OPEN_CONSOLE:
            sprintf(title, "%s: Import Console Font", app_name);
            s = XmStringCreateSimple("*");
            break;
          case XMBDFED_OPEN_PKGF:
            sprintf(title, "%s: Import PK/GF Font", app_name);
            s = XmStringCreateSimple("*[PpGg][KkFf]");
            break;
        }
        XtSetArg(av[0], XmNtitle, title);
        XtSetArg(av[1], XmNpattern, s);
        opend = XmCreateFileSelectionDialog(top, "xmbdfed_open_file_sel_box",
                                            av, 2);
        XmStringFree(s);
        XtAddCallback(opend, XmNokCallback, ReallyOpen, 0);
        XtAddCallback(opend, XmNcancelCallback, DoClose,
                      (XtPointer) XtParent(opend));
        XtManageChild(opend);
    } else {
        /*
         * Update the title and file match pattern based on the
         * open type.
         */
        switch (ed->filemenu.open_type) {
          case XMBDFED_OPEN_BDF:
            sprintf(title, "%s: Open BDF Font", app_name);
            s = XmStringCreateSimple("*.[Bb][Dd][Ff]");
            break;
          case XMBDFED_OPEN_HBF:
            sprintf(title, "%s: Import HBF Font", app_name);
            s = XmStringCreateSimple("*.[Hh][Bb][Ff]");
            break;
          case XMBDFED_OPEN_CONSOLE:
            sprintf(title, "%s: Import Console Font", app_name);
            s = XmStringCreateSimple("*");
            break;
          case XMBDFED_OPEN_PKGF:
            sprintf(title, "%s: Import PK/GF Font", app_name);
            s = XmStringCreateSimple("*[PpGg][KkFf]");
            break;
        }
        XtSetArg(av[0], XmNtitle, title);
        XtSetValues(XtParent(opend), av, 1);

        XtSetArg(av[0], XmNpattern, s);
        XtSetValues(opend, av, 1);
        XmStringFree(s);
    }
    active_editor = ed->id;
    XtPopup(XtParent(opend), XtGrabNone);
}

static void
#ifndef _NO_PROTO
DoOpen(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoOpen(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];
    ed->filemenu.open_type = XMBDFED_OPEN_BDF;
    OpenImportDialog(w, ed);
}

static void
#ifndef _NO_PROTO
ImportHBF(Widget w, XtPointer client_data, XtPointer call_data)
#else
ImportHBF(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];
    ed->filemenu.open_type = XMBDFED_OPEN_HBF;
    OpenImportDialog(w, ed);
}

static void
#ifndef _NO_PROTO
ImportConsole(Widget w, XtPointer client_data, XtPointer call_data)
#else
ImportConsole(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];
    ed->filemenu.open_type = XMBDFED_OPEN_CONSOLE;
    OpenImportDialog(w, ed);
}

static void
#ifndef _NO_PROTO
ImportPkGf(Widget w, XtPointer client_data, XtPointer call_data)
#else
ImportPkGf(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];
    ed->filemenu.open_type = XMBDFED_OPEN_PKGF;
    OpenImportDialog(w, ed);
}

static Boolean
#ifndef _NO_PROTO
ReallyQuit(XtPointer client_data)
#else
ReallyQuit(client_data)
XtPointer client_data;
#endif
{
    XtRemoveWorkProc(wpid);
    XtDestroyApplicationContext(app);

    bdf_cleanup();

    exit(0);

    return True;
}

static void
#ifndef _NO_PROTO
DoQuit(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoQuit(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    unsigned long i, cnt;
    MXFEditor *ed;

    exiting = True;

    /*
     * Go through and find out if any of the fonts have been modified.  Close
     * all open glyph editors at the same time which allows modified glyphs to
     * be updated.
     */
    for (i = cnt = 0, ed = editors; i < num_editors; i++, ed++) {
        GlyphEditClose(ed);
        if (XmuttFontGridModified(ed->fgrid) == True)
          cnt++;
    }

    if (cnt > 0) {
        if (cnt == 1) {
            if (num_editors > 1)
              sprintf(name, "A font has been modified.  Save?");
            else
              sprintf(name, "The font has been modified.  Save?");
        } else
          sprintf(name, "Some fonts have been modified.  Save?");
        if (AskBooleanQuestion(name, "Save And Exit",
                               "Exit Without Save") == True) {
            for (i = 0, ed = editors; i < num_editors; i++, ed++) {
                if (XmuttFontGridModified(ed->fgrid) == True) {
                    /*
                     * Make sure to only call DoSave() if the font has an
                     * associated file.
                     */
                    if (cnt == 1 && ed->file != 0)
                      DoSave(w, (XtPointer) ed->id, 0);
                    else
                      WaitSave(w, ed);
                }
            }
        }
    }

    /*
     * One of the save operations failed so we can't exit yet.
     */
    if (exiting == False)
      return;

    /*
     * Verify an exit is really wanted.
     */
    if (xmbdfed_opts.really_exit &&
        AskBooleanQuestion("Really Exit?", "Yes", "No") == False)
      return;

    /*
     * Go through and free up all the fonts.
     */
    for (i = 0, ed = editors; i < num_editors; i++, ed++) {
        if (ed->path != 0)
          XtFree(ed->path);
        if (ed->file != 0)
          XtFree(ed->file);
    }

    if (num_editors > 0)
      XtFree((char *) editors);
    num_editors = 0;

    /*
     * Clean up any glyph editors that were created.
     */
    GlyphEditCleanup();

    XtDestroyWidget(top);
    wpid = XtAppAddWorkProc(app, ReallyQuit, 0);
}

static void
#ifndef _NO_PROTO
PageChangeCallback(Widget w, XtPointer client_data, XtPointer call_data)
#else
PageChangeCallback(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    unsigned long id;
    MXFEditor *ed;
    XmuttFontGridPageInfoStruct *pi;

    id = (unsigned long) client_data;
    ed = &editors[id];
    pi = (XmuttFontGridPageInfoStruct *) call_data;

    if (pi->previous_page == -1) {
        XtSetSensitive(ed->prevpage, False);
        XtSetSensitive(ed->firstpage, False);
    } else {
        XtSetSensitive(ed->prevpage, True);
        XtSetSensitive(ed->firstpage, True);
    }

    if (pi->next_page == -1) {
        XtSetSensitive(ed->nextpage, False);
        XtSetSensitive(ed->lastpage, False);
    } else {
        XtSetSensitive(ed->nextpage, True);
        XtSetSensitive(ed->lastpage, True);
    }

    /*
     * Set the page number in the text field.
     */
    sprintf(name, "%ld", pi->current_page);
    XmTextFieldSetString(ed->pageno, name);
    XmTextFieldSetCursorPosition(ed->pageno, strlen(name));
}

static void
#ifndef _NO_PROTO
UpdateGlyphInfo(Widget w, XtPointer client_data, XtPointer call_data)
#else
UpdateGlyphInfo(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;
    short as, ds, rt, lt;
    unsigned int b1, b2, b3, b4;
    XmuttFontGridSelectionStruct *cb;
    XmString s;
    Arg av[1];

    ed = &editors[(unsigned long) client_data];
    cb = (XmuttFontGridSelectionStruct *) call_data;

    b1 = (cb->start >> 8) & 0xff;
    b2 = cb->start & 0xff;

    if (cb->start == cb->end) {
        as = ds = rt = lt = 0;
        if (cb->num_glyphs != 0 && cb->glyphs != 0) {
            as = cb->glyphs->bbx.ascent;
            ds = cb->glyphs->bbx.descent;
            rt = cb->glyphs->bbx.width + cb->glyphs->bbx.x_offset;
            lt = cb->glyphs->bbx.x_offset;
            if (cb->glyphs->name != 0)
              (void) strcpy(name, cb->glyphs->name);
            else
              sprintf(name, "char%ld", cb->glyphs->encoding);

            /*
             * If the glyph test dialog is active, send it the glyph.
             */
            GlyphTestAddGlyph(XmuttFontGridFont(ed->fgrid), cb->glyphs);
        } else
          sprintf(name, "char%ld", cb->start);

        switch (cb->base) {
          case 8:
            sprintf(title, OGINFO, name, cb->start, b1, b2, as, ds, rt, lt);
            break;
          case 10:
            sprintf(title, DGINFO, name, cb->start, b1, b2, as, ds, rt, lt);
            break;
          case 16:
            sprintf(title, HGINFO, name, cb->start, b1, b2, as, ds, rt, lt);
            break;
        }
    } else {
        b3 = (cb->end >> 8) & 0xff;
        b4 = cb->end & 0xff;
        switch (cb->base) {
          case 8:
            sprintf(title, "Selection\n%lo (%o,%o) - %lo (%o,%o)",
                    cb->start, b1, b2, cb->end, b3, b4);
            break;
          case 10:
            sprintf(title, "Selection\n%ld (%d,%d) - %ld (%d,%d)",
                    cb->start, b1, b2, cb->end, b3, b4);
            break;
          case 16:
            sprintf(title, "Selection\n%04lX (%02X,%02X) - %04lX (%02X,%02X)",
                    cb->start, b1, b2, cb->end, b3, b4);
            break;
        }
    }
    s = XmStringCreateLtoR(title, XmSTRING_DEFAULT_CHARSET);
    XtSetArg(av[0], XmNlabelString, s);
    XtSetValues(ed->glyphinfo, av, 1);
    XmStringFree(s);
}

static void
#ifndef _NO_PROTO
UpdateModified(Widget w, XtPointer client_data, XtPointer call_data)
#else
UpdateModified(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;
    XmuttFontGridModifiedCallbackStruct *cb;
    Arg av[1];

    ed = &editors[(unsigned long) client_data];
    cb = (XmuttFontGridModifiedCallbackStruct *) call_data;

    if (XmuttFontGridModified(ed->fgrid) == True) {
        if (ed->file == 0)
          sprintf(title, "%s - (unnamed%ld) [modified]", app_name, ed->id);
        else
          sprintf(title, "%s - %s [modified]", app_name, ed->file);
        XtSetSensitive(ed->filemenu.save, True);
    } else {
        if (ed->file == 0)
          sprintf(title, "%s - (unnamed%ld)", app_name, ed->id);
        else
          sprintf(title, "%s - %s", app_name, ed->file);
        XtSetSensitive(ed->filemenu.save, False);
    }

    XtSetArg(av[0], XmNtitle, title);
    XtSetValues(ed->shell, av, 1);

    /*
     * Check to see if the font name needs to be changed in the font name
     * text field.  Inhibit updates by the routine that tracks when the
     * font name text field changes.
     */
    if (ed->update_fontname == True && cb->reason == XmuttFG_NAME_CHANGE) {
        ed->update_fontname = False;
        XmTextFieldSetString(ed->fontname, cb->font_name);
        ed->update_fontname = True;
    }

    if (cb->reason == XmuttFG_RESIZE)
      /*
       * On resizes, the glyph test dialog needs to have the bounding
       * box updated.
       */
      GlyphTestResize();

    /*
     * If the callback was due to a replace action, then alert the glyph test
     * dialog if it is active.
     */
    if (cb->reason == XmuttFG_REPLACE)
      GlyphTestRedisplay();
}

static void
#ifndef _NO_PROTO
GotoCode(Widget w, XtPointer client_data, XtPointer call_data)
#else
GotoCode(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    long n;
    MXFEditor *ed;
    char *val;

    ed = &editors[(unsigned long) client_data];

    val = XmTextFieldGetString(ed->code);
    switch(XmuttFontGridCodeBase(ed->fgrid)) {
      case 8: sscanf(val, "%lo", &n); break;
      case 10: sscanf(val, "%ld", &n); break;
      case 16: sscanf(val, "%lx", &n); break;
    }
    XtFree(val);
    XmuttFontGridGotoCodePage(ed->fgrid, n);
}

static void
#ifndef _NO_PROTO
GotoPage(Widget w, XtPointer client_data, XtPointer call_data)
#else
GotoPage(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    long n;
    MXFEditor *ed;
    char *val;

    ed = &editors[(unsigned long) client_data];

    val = XmTextFieldGetString(ed->pageno);
    n = atol(val);
    XtFree(val);
    XmuttFontGridGotoPage(ed->fgrid, n);
}

static void
#ifndef _NO_PROTO
GotoFirstPage(Widget w, XtPointer client_data, XtPointer call_data)
#else
GotoFirstPage(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];

    XmuttFontGridFirstPage(ed->fgrid);
}

static void
#ifndef _NO_PROTO
GotoNextPage(Widget w, XtPointer client_data, XtPointer call_data)
#else
GotoNextPage(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    unsigned long id;
    MXFEditor *ed;

    id = (unsigned long) client_data;
    ed = &editors[id];

    XmuttFontGridNextPage(ed->fgrid);
}

static void
#ifndef _NO_PROTO
GotoPrevPage(Widget w, XtPointer client_data, XtPointer call_data)
#else
GotoPrevPage(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    unsigned long id;
    MXFEditor *ed;

    id = (unsigned long) client_data;
    ed = &editors[id];

    XmuttFontGridPreviousPage(ed->fgrid);
}

static void
#ifndef _NO_PROTO
GotoLastPage(Widget w, XtPointer client_data, XtPointer call_data)
#else
GotoLastPage(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    unsigned long id;
    MXFEditor *ed;

    id = (unsigned long) client_data;
    ed = &editors[id];

    XmuttFontGridLastPage(ed->fgrid);
}

static void
#ifndef _NO_PROTO
DisplayOctal(Widget w, XtPointer client_data, XtPointer call_data)
#else
DisplayOctal(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    unsigned int s1, s2, s3, s4;
    short as, ds, rt, lt;
    MXFEditor *ed;
    XmToggleButtonCallbackStruct *cb;
    XmString s;
    Arg av[1];
    XmuttFontGridSelectionStruct sinfo;

    ed = &editors[(unsigned long) client_data];
    cb = (XmToggleButtonCallbackStruct *) call_data;

    if (cb->set) {
        XmuttFontGridChangeCodeBase(ed->fgrid, 8);
        XmuttFontGridSelectionInfo(ed->fgrid, &sinfo);
        if (sinfo.start != -1 && sinfo.end != -1) {
            if (sinfo.start == sinfo.end) {
                s1 = (sinfo.start >> 8) & 0xff;
                s2 = sinfo.start & 0xff;
                as = ds = rt = lt = 0;
                if (sinfo.num_glyphs != 0) {
                    as = sinfo.glyphs->bbx.ascent;
                    ds = sinfo.glyphs->bbx.descent;
                    rt = sinfo.glyphs->bbx.width + sinfo.glyphs->bbx.x_offset;
                    lt = sinfo.glyphs->bbx.x_offset;
                    if (sinfo.glyphs->name != 0)
                      (void) strcpy(name, sinfo.glyphs->name);
                    else
                      sprintf(name, "char%ld", sinfo.glyphs->encoding);
                } else
                  sprintf(name, "char%ld", sinfo.start);
                sprintf(title, OGINFO, name, sinfo.start,
                        s1, s2, as, ds, rt, lt);
            } else {
                s3 = (sinfo.end >> 8) & 0xff;
                s4 = sinfo.end & 0xff;
                sprintf(title, "%lo (%o,%o) - %lo (%o,%o)",
                        sinfo.start, s1, s2, sinfo.end, s3, s4);
            }
            s = XmStringCreateSimple(title);
            XtSetArg(av[0], XmNlabelString, s);
            XtSetValues(ed->glyphinfo, av, 1);
            XmStringFree(s);
        }

        /*
         * Update the glyph editors.
         */
        GlyphEditChangeBase(ed->id, 8);
    }
}

static void
#ifndef _NO_PROTO
DisplayDecimal(Widget w, XtPointer client_data, XtPointer call_data)
#else
DisplayDecimal(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    unsigned int s1, s2, s3, s4;
    short as, ds, rt, lt;
    MXFEditor *ed;
    XmToggleButtonCallbackStruct *cb;
    XmString s;
    Arg av[1];
    XmuttFontGridSelectionStruct sinfo;

    ed = &editors[(unsigned long) client_data];
    cb = (XmToggleButtonCallbackStruct *) call_data;

    if (cb->set) {
        XmuttFontGridChangeCodeBase(ed->fgrid, 10);
        XmuttFontGridSelectionInfo(ed->fgrid, &sinfo);
        if (sinfo.start != -1 && sinfo.end != -1) {
            if (sinfo.start == sinfo.end) {
                s1 = (sinfo.start >> 8) & 0xff;
                s2 = sinfo.start & 0xff;
                as = ds = rt = lt = 0;
                if (sinfo.num_glyphs != 0) {
                    as = sinfo.glyphs->bbx.ascent;
                    ds = sinfo.glyphs->bbx.descent;
                    rt = sinfo.glyphs->bbx.width + sinfo.glyphs->bbx.x_offset;
                    lt = sinfo.glyphs->bbx.x_offset;
                    if (sinfo.glyphs->name != 0)
                      (void) strcpy(name, sinfo.glyphs->name);
                    else
                      sprintf(name, "char%ld", sinfo.glyphs->encoding);
                } else
                  sprintf(name, "char%ld", sinfo.start);
                sprintf(title, DGINFO, name, sinfo.start, s1, s2,
                        as, ds, rt, lt);
            } else {
                s3 = (sinfo.end >> 8) & 0xff;
                s4 = sinfo.end & 0xff;
                sprintf(title, "%ld (%d,%d) - %ld (%d,%d)",
                        sinfo.start, s1, s2, sinfo.end, s3, s4);
            }
            s = XmStringCreateLtoR(title, XmSTRING_DEFAULT_CHARSET);
            XtSetArg(av[0], XmNlabelString, s);
            XtSetValues(ed->glyphinfo, av, 1);
            XmStringFree(s);
        }

        /*
         * Update the glyph editors.
         */
        GlyphEditChangeBase(ed->id, 10);
    }
}

static void
#ifndef _NO_PROTO
DisplayHex(Widget w, XtPointer client_data, XtPointer call_data)
#else
DisplayHex(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    unsigned int s1, s2, s3, s4;
    short as, ds, rt, lt;
    MXFEditor *ed;
    XmToggleButtonCallbackStruct *cb;
    XmString s;
    Arg av[1];
    XmuttFontGridSelectionStruct sinfo;

    ed = &editors[(unsigned long) client_data];
    cb = (XmToggleButtonCallbackStruct *) call_data;

    if (cb->set) {
        XmuttFontGridChangeCodeBase(ed->fgrid, 16);
        XmuttFontGridSelectionInfo(ed->fgrid, &sinfo);
        if (sinfo.start != -1 && sinfo.end != -1) {
            if (sinfo.start == sinfo.end) {
                s1 = (sinfo.start >> 8) & 0xff;
                s2 = sinfo.start & 0xff;
                as = ds = rt = lt = 0;
                if (sinfo.num_glyphs != 0) {
                    as = sinfo.glyphs->bbx.ascent;
                    ds = sinfo.glyphs->bbx.descent;
                    rt = sinfo.glyphs->bbx.width + sinfo.glyphs->bbx.x_offset;
                    lt = sinfo.glyphs->bbx.x_offset;
                    if (sinfo.glyphs->name != 0)
                      (void) strcpy(name, sinfo.glyphs->name);
                    else
                      sprintf(name, "char%ld", sinfo.glyphs->encoding);
                } else
                  sprintf(name, "char%ld", sinfo.start);
                sprintf(title, HGINFO, name, sinfo.start, s1, s2,
                        as, ds, rt, lt);
            } else {
                s3 = (sinfo.end >> 8) & 0xff;
                s4 = sinfo.end & 0xff;
                sprintf(title, "%04lX (%02X,%02X) - %04lX (%02X,%02X)",
                        sinfo.start, s1, s2, sinfo.end, s3, s4);
            }
            s = XmStringCreateSimple(title);
            XtSetArg(av[0], XmNlabelString, s);
            XtSetValues(ed->glyphinfo, av, 1);
            XmStringFree(s);
        }

        /*
         * Update the glyph editors.
         */
        GlyphEditChangeBase(ed->id, 16);
    }
}

static void
#ifndef _NO_PROTO
ToggleEncodedView(Widget w, XtPointer client_data, XtPointer call_data)
#else
ToggleEncodedView(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;
    Widget pb;
    XmString s;
    XmuttFontGridPageInfoStruct pi;

    ed = &editors[(unsigned long) client_data];

    pb = (Widget) client_data;    

    XmuttFontGridPageInfo(ed->fgrid, &pi);

    if (pi.unencoded == True) {
        s = XmStringCreateSimple("Unencoded");
        XmuttFontGridViewUnencoded(ed->fgrid, False);
    } else {
        s = XmStringCreateSimple("Encoded");
        XmuttFontGridViewUnencoded(ed->fgrid, True);
    }
    XtVaSetValues(ed->viewmenu.unencoded, XmNlabelString, s, 0);
    XmStringFree(s);
}

static void
#ifndef _NO_PROTO
ShowMessages(Widget w, XtPointer client_data, XtPointer call_data)
#else
ShowMessages(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;
    Widget frame, pb;
    char *s;
    Cardinal ac;
    Arg av[8];

    ed = &editors[(unsigned long) client_data];

    if (acmsg_shell == 0) {
        /*
         * Create the auto-correction message dialog.
         */
        ac = 0;
        XtSetArg(av[ac], XmNallowShellResize, True); ac++;
        XtSetArg(av[ac], XmNdeleteResponse, XmUNMAP); ac++;
        acmsg_shell = XtCreatePopupShell("xmbdfed_acmsgs_shell",
                                         xmDialogShellWidgetClass,
                                         top, av, ac);

        ac = 0;
        XtSetArg(av[ac], XmNverticalSpacing, 5); ac++;
        XtSetArg(av[ac], XmNmarginWidth, 2); ac++;
        XtSetArg(av[ac], XmNmarginHeight, 2); ac++;
        acmsg_form = XtCreateWidget("xmbdfed_acmsgs_form", xmFormWidgetClass,
                                    acmsg_shell, av, ac);

        ac = 0;
        XtSetArg(av[ac], XmNtopAttachment, XmATTACH_FORM); ac++;
        XtSetArg(av[ac], XmNleftAttachment, XmATTACH_FORM); ac++;
        XtSetArg(av[ac], XmNrightAttachment, XmATTACH_FORM); ac++;
        XtSetArg(av[ac], XmNcolumns, 40); ac++;
        XtSetArg(av[ac], XmNrows, 10); ac++;
        XtSetArg(av[ac], XmNeditMode, XmMULTI_LINE_EDIT); ac++;
        XtSetArg(av[ac], XmNeditable, False); ac++;
        XtSetArg(av[ac], XmNcursorPositionVisible, False); ac++;
        acmsg_text = XmCreateScrolledText(acmsg_form, "xmbdfed_acmsgs_text",
                                          av, ac);
        XtManageChild(acmsg_text);

        ac = 0;
        XtSetArg(av[ac], XmNtopAttachment, XmATTACH_WIDGET); ac++;
        XtSetArg(av[ac], XmNtopWidget, XtParent(acmsg_text)); ac++;
        XtSetArg(av[ac], XmNleftAttachment, XmATTACH_FORM); ac++;
        XtSetArg(av[ac], XmNrightAttachment, XmATTACH_FORM); ac++;
        XtSetArg(av[ac], XmNbottomAttachment, XmATTACH_FORM); ac++;
        frame = XtCreateManagedWidget("xmbdfed_acmsgs_frame",
                                      xmFrameWidgetClass, acmsg_form, av, ac);

        pb = XtCreateManagedWidget("Close", xmPushButtonWidgetClass, frame,
                                   0, 0);
        XtAddCallback(pb, XmNactivateCallback, DoClose,
                      (XtPointer) acmsg_shell);
    }

    /*
     * Adjust the title to match the editor.
     */
    if (ed->file == 0)
      sprintf(title, "%s - Corrections: (unnamed%ld)", app_name, ed->id);
    else
      sprintf(title, "%s - Corrections: %s", app_name, ed->file);

    XtSetArg(av[0], XmNtitle, title);
    XtSetValues(acmsg_shell, av, 1);

    /*
     * Replace the text with any auto-correction messages the font might have.
     */
    if ((s = XmuttFontGridFontMessages(ed->fgrid)) == 0)
      s = "";
    XmTextSetString(acmsg_text, s);

    XtManageChild(acmsg_form);
    XtPopup(acmsg_shell, XtGrabNone);
}

static void
#ifndef _NO_PROTO
DoFontGridCopy(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoFontGridCopy(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];
    XmuttFontGridCopySelection(ed->fgrid);
}

static void
#ifndef _NO_PROTO
DoFontGridCut(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoFontGridCut(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];
    XmuttFontGridCutSelection(ed->fgrid);
}

static void
#ifndef _NO_PROTO
DoFontGridPaste(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoFontGridPaste(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];
    XmuttFontGridPasteSelection(ed->fgrid);
}

static void
#ifndef _NO_PROTO
MakeXLFDName(Widget w, XtPointer client_data, XtPointer call_data)
#else
MakeXLFDName(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];
    XmuttFontGridMakeXLFDName(ed->fgrid);
}

static void
#ifndef _NO_PROTO
DoUpdateNameFromProps(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoUpdateNameFromProps(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];
    XmuttFontGridUpdateNameFromProperties(ed->fgrid);
}

static void
#ifndef _NO_PROTO
DoUpdatePropsFromName(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoUpdatePropsFromName(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];
    XmuttFontGridUpdatePropertiesFromName(ed->fgrid);
}

static void
#ifndef _NO_PROTO
DoUpdateAverageWidth(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoUpdateAverageWidth(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];

    XmuttFontGridUpdateAverageWidth(ed->fgrid);
}

static void
#ifndef _NO_PROTO
DoFontNameChange(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoFontNameChange(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;
    char *s;

    ed = &editors[(unsigned long) client_data];

    if (ed->update_fontname == True) {
        /*
         * Tell the font modified routine to avoid replacing the font name
         * because it is changing due to user editing.
         */
        ed->update_fontname = False;

        s = XmTextFieldGetString(ed->fontname);
        XmuttFontGridSetFontName(ed->fgrid, s);
        free(s);

        ed->update_fontname = True;
    }
}

static void
#ifndef _NO_PROTO
DoSetMainEditMenu(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoSetMainEditMenu(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;
    XmuttFontGridPageInfoStruct pi;

    ed = &editors[(unsigned long) client_data];
    if (XmuttFontGridClipboardEmpty(ed->fgrid) == True)
      XtSetSensitive(ed->editmenu.paste, False);
    else
      XtSetSensitive(ed->editmenu.paste, True);

    if (XmuttFontGridHasSelection(ed->fgrid) == True) {
        XtSetSensitive(ed->editmenu.copy, True);
        XtSetSensitive(ed->editmenu.cut, True);
    } else {
        XtSetSensitive(ed->editmenu.copy, False);
        XtSetSensitive(ed->editmenu.cut, False);
    }

    XmuttFontGridPageInfo(ed->fgrid, &pi);
    if (pi.unencoded == True) {
        if (pi.num_unencoded > 0)
          XtSetSensitive(ed->editmenu.test, True);
        else
          XtSetSensitive(ed->editmenu.test, False);
    } else {
        if (pi.num_encoded > 0)
          XtSetSensitive(ed->editmenu.test, True);
        else
          XtSetSensitive(ed->editmenu.test, False);
    }

    if (XmuttFontGridHasXLFDName(ed->fgrid) == True) {
        XtSetSensitive(ed->editmenu.xlfdname, False);
        XtSetSensitive(ed->editmenu.unfp, True);
        XtSetSensitive(ed->editmenu.upfn, True);
        XtSetSensitive(ed->editmenu.awidth, True);
    } else {
        XtSetSensitive(ed->editmenu.xlfdname, True);
        XtSetSensitive(ed->editmenu.unfp, False);
        XtSetSensitive(ed->editmenu.upfn, False);
        XtSetSensitive(ed->editmenu.awidth, False);
    }
}

static void
#ifndef _NO_PROTO
DoResetMainEditMenu(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoResetMainEditMenu(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];
    XtSetSensitive(ed->editmenu.paste, True);
    XtSetSensitive(ed->editmenu.copy, True);
    XtSetSensitive(ed->editmenu.cut, True);
    XtSetSensitive(ed->editmenu.xlfdname, True);
}

static void
#ifndef _NO_PROTO
DoBuildEditorMenu(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoBuildEditorMenu(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    unsigned long id, i;
    int nkids;
    MXFEditor *ed;
    Widget pb, *kids;
    XmString s;
    Arg av[2];

    id = (unsigned long) client_data;
    ed = &editors[id];

    XtSetArg(av[0], XmNnumChildren, &nkids);
    XtSetArg(av[1], XmNchildren, &kids);
    XtGetValues(ed->elist, av, 2);

    nkids -= 2;

    /*
     * Update the existing menu entries for windows.
     */
    for (i = 0; i < nkids; i++) {
        if (editors[i].file == 0)
          sprintf(title, "(unnamed%ld)", i);
        else
          sprintf(title, "%s", editors[i].file);
        s = XmStringCreateSimple(title);
        XtSetArg(av[0], XmNlabelString, s);
        XtSetValues(kids[i + 2], av, 1);
        XmStringFree(s);
    }

    /*
     * Add any remaining editors that were created since the menu was last
     * shown.
     */
    for (; i < num_editors; i++) {
        if (editors[i].file == 0) {
            sprintf(name, "(unnamed%ld)", i);
            pb = XtCreateManagedWidget(name, xmPushButtonWidgetClass,
                                       ed->elist, 0, 0);
        } else
            pb = XtCreateManagedWidget(editors[i].file,
                                       xmPushButtonWidgetClass,
                                       ed->elist, 0, 0);
        XtAddCallback(pb, XmNactivateCallback, ShowEditor,
                      (XtPointer) editors[i].id);
    }
}

static void
#ifndef _NO_PROTO
DoNewEditor(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoNewEditor(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    unsigned long id;

    id = MakeEditor(top, 0, False);
    ShowEditor(w, (XtPointer) id, 0);
}

static void
#ifndef _NO_PROTO
MakeMenuBar(Widget mbar, MXFEditor *ed)
#else
MakeMenuBar(mbar, ed)
Widget mbar;
MXFEditor *ed;
#endif
{
    Widget pdown, cpdown, casc, pb, sep;

    XmString s;
    Cardinal ac;
    Arg av[6];

    sprintf(name, "xmbdfed_editor_%ld_file_menu", ed->id);
    pdown = XmCreatePulldownMenu(mbar, name, 0, 0);
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'F'); ac++;
    XtSetArg(av[ac], XmNsubMenuId, pdown); ac++;
    casc = XtCreateManagedWidget("File", xmCascadeButtonWidgetClass, mbar,
                                 av, ac);

    s = XmStringCreateSimple("Ctrl+N");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'N'); ac++;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>n"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    pb = XtCreateManagedWidget("New", xmPushButtonWidgetClass, pdown, av, ac);
    XtAddCallback(pb, XmNactivateCallback, DoNewEditor, 0);
    XmStringFree(s);

    s = XmStringCreateSimple("Ctrl+O");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'O'); ac++;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>o"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    pb = XtCreateManagedWidget("Open", xmPushButtonWidgetClass, pdown, av, ac);
    XtAddCallback(pb, XmNactivateCallback, DoOpen, (XtPointer) ed->id);
    XmStringFree(s);

    s = XmStringCreateSimple("Ctrl+S");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'S'); ac++;
    XtSetArg(av[ac], XmNsensitive, False); ac++;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>s"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    ed->filemenu.save = XtCreateManagedWidget("Save", xmPushButtonWidgetClass,
                                              pdown, av, ac);
    XtAddCallback(ed->filemenu.save, XmNactivateCallback, DoSave,
                  (XtPointer) ed->id);
    XmStringFree(s);

    s = XmStringCreateSimple("Ctrl+W");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'A'); ac++;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>w"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    ed->filemenu.saveas = XtCreateManagedWidget("Save As",
                                                xmPushButtonWidgetClass,
                                                pdown, av, ac);
    XtAddCallback(ed->filemenu.saveas, XmNactivateCallback, DoSaveAs,
                  (XtPointer) ed->id);
    XmStringFree(s);

    sprintf(name, "xmbdfed_editor_%ld_filesep1", ed->id);
    sep = XtCreateManagedWidget(name, xmSeparatorWidgetClass, pdown, 0, 0);

    sprintf(name, "xmbdfed_editor_%ld_import_menu", ed->id);
    cpdown = XmCreatePulldownMenu(pdown, name, 0, 0);

    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'I'); ac++;
    XtSetArg(av[ac], XmNsubMenuId, cpdown); ac++;
    pb = XtCreateManagedWidget("Import", xmCascadeButtonWidgetClass, pdown,
                               av, ac);

    s = XmStringCreateSimple("Ctrl+K");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'P'); ac++;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>k"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    pb = XtCreateManagedWidget("PK/GF Font", xmPushButtonWidgetClass,
                               cpdown, av, ac);
    XtAddCallback(pb, XmNactivateCallback, ImportPkGf, (XtPointer) ed->id);
    XmStringFree(s);

    s = XmStringCreateSimple("Ctrl+L");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'C'); ac++;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>l"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    pb = XtCreateManagedWidget("Console Font", xmPushButtonWidgetClass,
                               cpdown, av, ac);
    XtAddCallback(pb, XmNactivateCallback, ImportConsole, (XtPointer) ed->id);
    XmStringFree(s);

    s = XmStringCreateSimple("Ctrl+H");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'H'); ac++;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>h"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    pb = XtCreateManagedWidget("HBF Font", xmPushButtonWidgetClass, cpdown,
                               av, ac);
    XtAddCallback(pb, XmNactivateCallback, ImportHBF, (XtPointer) ed->id);
    XmStringFree(s);

    sprintf(name, "xmbdfed_editor_%ld_importsep", ed->id);
    sep = XtCreateManagedWidget(name, xmSeparatorWidgetClass, cpdown, 0, 0);

    s = XmStringCreateSimple("Ctrl+G");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'S'); ac++;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>g"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    pb = XtCreateManagedWidget("Server Font", xmPushButtonWidgetClass, cpdown,
                               av, ac);
    XtAddCallback(pb, XmNactivateCallback, DoGrabServerFont,
                  (XtPointer) ed->id);
    XmStringFree(s);

    sprintf(name, "xmbdfed_editor_%ld_filesep2", ed->id);
    sep = XtCreateManagedWidget(name, xmSeparatorWidgetClass, pdown, 0, 0);

    s = XmStringCreateSimple("Ctrl+F4");
    if (ed->id == 0) {
        ac = 0;
        XtSetArg(av[ac], XmNmnemonic, 'E'); ac++;
        XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>F4"); ac++;
        XtSetArg(av[ac], XmNacceleratorText, s); ac++;
        pb = XtCreateManagedWidget("Exit", xmPushButtonWidgetClass, pdown,
                                   av, ac);
        XtAddCallback(pb, XmNactivateCallback, DoQuit, 0);
    } else {
        ac = 0;
        XtSetArg(av[ac], XmNmnemonic, 'C');
        XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>F4"); ac++;
        XtSetArg(av[ac], XmNacceleratorText, s); ac++;
        pb = XtCreateManagedWidget("Close", xmPushButtonWidgetClass, pdown,
                                   av, ac);
        XtAddCallback(pb, XmNactivateCallback, DoCloseEditor,
                      (XtPointer) ed->id);
    }
    XmStringFree(s);

    sprintf(name, "xmbdfed_editor_%ld_edit_menu", ed->id);
    pdown = XmCreatePulldownMenu(mbar, name, 0, 0);
    XtAddCallback(pdown, XmNmapCallback, DoSetMainEditMenu,
                  (XtPointer) ed->id);
    XtAddCallback(pdown, XmNunmapCallback, DoResetMainEditMenu,
                  (XtPointer) ed->id);

    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'E'); ac++;
    XtSetArg(av[ac], XmNsubMenuId, pdown); ac++;
    casc = XtCreateManagedWidget("Edit", xmCascadeButtonWidgetClass, mbar,
                                 av, ac);

    s = XmStringCreateSimple("Ctrl+C");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'C'); ac++;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>c"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    ed->editmenu.copy = XtCreateManagedWidget("Copy", xmPushButtonWidgetClass,
                                              pdown, av, ac);
    XtAddCallback(ed->editmenu.copy, XmNactivateCallback, DoFontGridCopy,
                  (XtPointer) ed->id);
    XmStringFree(s);

    s = XmStringCreateSimple("Ctrl+X");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'u'); ac++;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>x"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    ed->editmenu.cut = XtCreateManagedWidget("Cut", xmPushButtonWidgetClass,
                                             pdown, av, ac);
    XtAddCallback(ed->editmenu.cut, XmNactivateCallback, DoFontGridCut,
                  (XtPointer) ed->id);
    XmStringFree(s);

    s = XmStringCreateSimple("Ctrl+V");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'P'); ac++;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>v"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    ed->editmenu.paste = XtCreateManagedWidget("Paste",
                                               xmPushButtonWidgetClass, pdown,
                                               av, ac);
    XtAddCallback(ed->editmenu.paste, XmNactivateCallback, DoFontGridPaste,
                  (XtPointer) ed->id);
    XmStringFree(s);

    sprintf(name, "xmbdfed_editor_%ld_editsep1", ed->id);
    sep = XtCreateManagedWidget(name, xmSeparatorWidgetClass, pdown, 0, 0);

    s = XmStringCreateSimple("Ctrl+P");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'r'); ac++;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>p"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    pb = XtCreateManagedWidget("Properties", xmPushButtonWidgetClass, pdown,
                               av, ac);
    XtAddCallback(pb, XmNactivateCallback, DoFontProperties,
                  (XtPointer) ed->id);
    XmStringFree(s);

    s = XmStringCreateSimple("Ctrl+M");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'm'); ac++;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>m"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    pb = XtCreateManagedWidget("Comments", xmPushButtonWidgetClass, pdown,
                               av, ac);
    XtAddCallback(pb, XmNactivateCallback, DoEditComments,
                  (XtPointer) ed->id);
    XmStringFree(s);

    sprintf(name, "xmbdfed_editor_%ld_editsep2", ed->id);
    sep = XtCreateManagedWidget(name, xmSeparatorWidgetClass, pdown, 0, 0);

    s = XmStringCreateSimple("Ctrl+I");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'F'); ac++;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>i"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    ed->editmenu.finfo = XtCreateManagedWidget("Font Info",
                                               xmPushButtonWidgetClass, pdown,
                                               av, ac);
    XtAddCallback(ed->editmenu.finfo, XmNactivateCallback,
                  DoFontInfo, (XtPointer) ed->id);
    XmStringFree(s);

    sprintf(name, "xmbdfed_editor_%ld_font_name_menu", ed->id);
    XtSetArg(av[0], XmNradioBehavior, True);
    cpdown = XmCreatePulldownMenu(pdown, name, av, 1);

    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'N'); ac++;
    XtSetArg(av[ac], XmNsubMenuId, cpdown); ac++;
    casc = XtCreateManagedWidget("Font Name", xmCascadeButtonWidgetClass,
                                 pdown, av, ac);

    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'X'); ac++;
    ed->editmenu.xlfdname = XtCreateManagedWidget("Make XLFD Name",
                                                  xmPushButtonWidgetClass,
                                                  cpdown, av, ac);
    XtAddCallback(ed->editmenu.xlfdname, XmNactivateCallback, MakeXLFDName,
                  (XtPointer) ed->id);

    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'N'); ac++;
    ed->editmenu.unfp = XtCreateManagedWidget("Update Name From Properties",
                                              xmPushButtonWidgetClass, cpdown,
                                              av, ac);
    XtAddCallback(ed->editmenu.unfp, XmNactivateCallback,
                  DoUpdateNameFromProps, (XtPointer) ed->id);

    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'P'); ac++;
    ed->editmenu.upfn = XtCreateManagedWidget("Update Properties From Name",
                                              xmPushButtonWidgetClass, cpdown,
                                              av, ac);
    XtAddCallback(ed->editmenu.upfn, XmNactivateCallback,
                  DoUpdatePropsFromName, (XtPointer) ed->id);

    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'A'); ac++;
    ed->editmenu.awidth = XtCreateManagedWidget("Update Average Width",
                                                xmPushButtonWidgetClass,
                                                cpdown, av, ac);
    XtAddCallback(ed->editmenu.awidth, XmNactivateCallback,
                  DoUpdateAverageWidth, (XtPointer) ed->id);

    sprintf(name, "xmbdfed_editor_%ld_editsep3", ed->id);
    sep = XtCreateManagedWidget(name, xmSeparatorWidgetClass, pdown, 0, 0);

    s = XmStringCreateSimple("Ctrl+Z");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'T'); ac++;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>z"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    ed->editmenu.test = XtCreateManagedWidget("Test Glyphs",
                                              xmPushButtonWidgetClass, pdown,
                                              av, ac);
    XmStringFree(s);
    XtAddCallback(ed->editmenu.test, XmNactivateCallback, DoGlyphTest,
                  (XtPointer) ed->id);

    sprintf(name, "xmbdfed_editor_%ld_editsep4", ed->id);
    sep = XtCreateManagedWidget(name, xmSeparatorWidgetClass, pdown, 0, 0);

    s = XmStringCreateSimple("Ctrl+T");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'S'); ac++;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>t"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    pb = XtCreateManagedWidget("Setup", xmPushButtonWidgetClass, pdown,
                               av, ac);
    XtAddCallback(pb, XmNactivateCallback, DoSetupDialog, (XtPointer) ed->id);
    XmStringFree(s);

    sprintf(name, "xmbdfed_editor_%ld_view_menu", ed->id);
    pdown = XmCreatePulldownMenu(mbar, name, 0, 0);
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'V'); ac++;
    XtSetArg(av[ac], XmNsubMenuId, pdown); ac++;
    casc = XtCreateManagedWidget("View", xmCascadeButtonWidgetClass, mbar,
                                 av, ac);

    s = XmStringCreateSimple("Ctrl+E");
    ac = 0;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>e"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    ed->viewmenu.unencoded = XtCreateManagedWidget("Unencoded",
                                                   xmPushButtonWidgetClass,
                                                   pdown, av, ac);
    XtAddCallback(ed->viewmenu.unencoded, XmNactivateCallback,
                  ToggleEncodedView, (XtPointer) ed->id);
    XmStringFree(s);
                                 
    sprintf(name, "xmbdfed_editor_%ld_view_code_menu", ed->id);
    XtSetArg(av[0], XmNradioBehavior, True);
    cpdown = XmCreatePulldownMenu(pdown, name, av, 1);

    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'C'); ac++;
    XtSetArg(av[ac], XmNsubMenuId, cpdown); ac++;
    casc = XtCreateManagedWidget("Code Base", xmCascadeButtonWidgetClass,
                                 pdown, av, ac);

    ed->viewmenu.oct = XtCreateManagedWidget("Octal",
                                             xmToggleButtonWidgetClass,
                                             cpdown, 0, 0);
    XtAddCallback(ed->viewmenu.oct, XmNvalueChangedCallback, DisplayOctal,
                  (XtPointer) ed->id);

    ed->viewmenu.dec = XtCreateManagedWidget("Decimal",
                                             xmToggleButtonWidgetClass,
                                             cpdown, 0, 0);
    XtAddCallback(ed->viewmenu.dec, XmNvalueChangedCallback, DisplayDecimal,
                  (XtPointer) ed->id);
    ed->viewmenu.hex = XtCreateManagedWidget("Hexadecimal",
                                             xmToggleButtonWidgetClass, cpdown,
                                             0, 0);
    XtAddCallback(ed->viewmenu.hex, XmNvalueChangedCallback, DisplayHex,
                  (XtPointer) ed->id);

    sprintf(name, "xmbdfed_editor_%ld_viewsep1", ed->id);
    sep = XtCreateManagedWidget(name, xmSeparatorWidgetClass, pdown, 0, 0);

    s = XmStringCreateSimple("Ctrl+A");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'M'); ac++;
    XtSetArg(av[ac], XmNaccelerator, "Ctrl<Key>a"); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    ed->viewmenu.acmsgs = XtCreateManagedWidget("Messages",
                                                xmPushButtonWidgetClass,
                                                pdown, av, ac);
    XtAddCallback(ed->viewmenu.acmsgs, XmNactivateCallback,
                  ShowMessages, (XtPointer) ed->id);
    XmStringFree(s);

    sprintf(name, "xmbdfed_editor_%ld_editor_menu", ed->id);
    ed->elist = pdown = XmCreatePulldownMenu(mbar, name, 0, 0);
    XtAddCallback(XtParent(pdown), XmNpopupCallback, DoBuildEditorMenu,
                  (XtPointer) ed->id);
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'd'); ac++;
    XtSetArg(av[ac], XmNsubMenuId, pdown); ac++;
    casc = XtCreateManagedWidget("Editors", xmCascadeButtonWidgetClass, mbar,
                                 av, ac);

    s = XmStringCreateSimple("Ctrl+N");
    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'N'); ac++;
    XtSetArg(av[ac], XmNacceleratorText, s); ac++;
    pb = XtCreateManagedWidget("New", xmPushButtonWidgetClass, pdown, av, ac);
    XtAddCallback(pb, XmNactivateCallback, DoNewEditor, 0);
    XmStringFree(s);

    sprintf(name, "xmbdfed_editor_%ld_editorsep1", ed->id);
    sep = XtCreateManagedWidget(name, xmSeparatorWidgetClass, pdown, 0, 0);

    sprintf(name, "xmbdfed_editor_%ld_help_menu", ed->id);
    pdown = XmCreatePulldownMenu(mbar, name, 0, 0);

    ac = 0;
    XtSetArg(av[ac], XmNmnemonic, 'H'); ac++;
    XtSetArg(av[ac], XmNsubMenuId, pdown); ac++;
    casc = XtCreateManagedWidget("Help", xmCascadeButtonWidgetClass, mbar,
                                 av, ac);

    /*
     * Tell the menu bar that this is the Help menu.
     */
    ac = 0;
    XtSetArg(av[ac], XmNmenuHelpWidget, casc); ac++;
    XtSetValues(mbar, av, ac);

    /*
     * Add the help menu to the pulldown.
     */
    AddHelpMenu(pdown);
}

static unsigned long
#ifndef _NO_PROTO
MakeEditor(Widget parent, char *filename, Boolean cmdline)
#else
MakeEditor(parent, filename, cmdline)
Widget parent;
char *filename;
Boolean cmdline;
#endif
{
    int hbf, fnamelen;
    Widget mw, mb, rc, rc1, frame, form, label1, label2;
    FILE *in;
    char *path, *file, *cp;
    unsigned long which;
    XmString s;
    bdf_font_t *font;
    bdf_callback_t callback;
    MXFEditor *ed;
    Cardinal ac;
    Dimension wd, ht;
    Boolean free_filename;
    XmuttFontGridPageInfoStruct pi;
    Arg av[12];

    which = num_editors;
    font = 0;
    path = file = 0;
    free_filename = False;

    callback = (cmdline == False) ? UpdateProgressBar : 0;

    /*
     * If a file has been passed, load it before creating the
     * editor.
     */
    if (filename != 0) {
        /*
         * Check to see if the font being loaded is an HBF or a BDF font.
         */
        if ((hbf = IsHBFFont(filename, 0)) < 0) {
            if (which > 0)
              return ~0L;
        }

        /*
         * Get the filename to be used for the progress bar.
         */
        if ((cp = strrchr(filename, '/')) == 0)
          cp = filename;
        else
          cp++;

        if (hbf > 0) {
            if ((font = bdf_load_hbf_font(filename,
                                          &xmbdfed_opts.font_opts,
                                          callback, cp)) == 0) {
                fprintf(stderr, "%s: unable to open HBF font '%s'.\n",
                        app_name, filename);
                if (which > 0)
                  return ~0L;
            } else {
                /*
                 * If the path name has a ".hbf" extension, replace it with
                 * ".bdf" so the font filename will have a useful extension.
                 * If there is no extension, assume the name needs to have a
                 * ".bdf" extension added so the HBF file will not be
                 * overwritten.
                 */
                free_filename = True;
                fnamelen = strlen(filename);
                path = (char *) XtMalloc(fnamelen + 1);
                (void) memcpy(path, filename, fnamelen + 1);
                filename = path;
                if ((file = strrchr(filename, '.')) != 0) {
                    file++;
                    if (*file == 'h' || *file == 'H')
                      (void) memcpy(file, "bdf", 3);
                } else {
                    filename = (char *) XtRealloc(filename, fnamelen + 5);
                    file = filename + fnamelen + 1;
                    /*
                     * Use sprintf() to make sure the path is NULL terminated.
                     */
                    sprintf(file, ".bdf");
                }
            }
        } else {
            in = fopen(filename, "r");
            if (in == 0) {
                fprintf(stderr, "%s: unable to to open BDF font '%s'.\n",
                        app_name, filename);
                if (which > 0)
                  return ~0L;
            } else {
                font = bdf_load_font(in, &xmbdfed_opts.font_opts,
                                     callback, cp);
                fclose(in);
                if (font == 0)
                  fprintf(stderr, "%s: problem loading BDF font '%s'.\n",
                          app_name, filename);
            }
        }

        file = path = 0;

        if (font != 0) {
            /*
             * Construct the path and filename from the one passed.
             */
            file = strrchr(filename, '/');
            if (file == 0) {
                path = 0;
                file = (char *) XtMalloc(strlen(filename) + 1);
                (void) strcpy(file, filename);
            } else {
                path = (char *) XtMalloc((file - filename) + 1);
                (void) strncpy(path, filename, (file - filename));
                path[file - filename] = 0;
                cp = ++file;
                file = (char *) XtMalloc(strlen(file) + 1);
                (void) strcpy(file, cp);
            }
        }

        /*
         * Free up the filename if indicated.
         */
        if (free_filename == True)
          XtFree(filename);
    }

    if (font == 0)
      sprintf(title, "%s - (unnamed%ld)", app_name, which);
    else
      sprintf(title, "%s - %s", app_name, file);

    if (which == 0)
      editors = (MXFEditor *) XtMalloc(sizeof(MXFEditor));
    else
      editors = (MXFEditor *) XtRealloc((char *) editors,
                                        sizeof(MXFEditor) * (num_editors + 1));

    ed = &editors[num_editors++];
    (void) memset((char *) ed, 0, sizeof(MXFEditor));

    ed->id = which;
    ed->path = path;
    ed->file = file;
    ed->update_fontname = True;

    sprintf(name, "xmbdfed_editor_%ld", which);
    ac = 0;
    XtSetArg(av[ac], XmNtitle, title); ac++;
    XtSetArg(av[ac], XmNallowShellResize, True); ac++;
    if (which == 0) {
        ed->shell = parent;
        XtSetValues(parent, av, 1);
    } else {
        XtSetArg(av[ac], XmNdeleteResponse, XmUNMAP); ac++;
        ed->shell = XtCreatePopupShell(name, topLevelShellWidgetClass, parent,
                                       av, ac);
    }

    sprintf(name, "xmbdfed_editor_%ld_mainw", which);
    ed->mainw = mw = XtCreateManagedWidget(name, xmMainWindowWidgetClass,
                                           ed->shell, 0, 0);

    sprintf(name, "xmbdfed_editor_%ld_menubar", which);
    mb = XmCreateMenuBar(mw, name, 0, 0);
    MakeMenuBar(mb, ed);
    XtManageChild(mb);

    /*
     * The primary RowColumn widget.
     */
    ac = 0;
    XtSetArg(av[ac], XmNverticalSpacing, 4); ac++;
    XtSetArg(av[ac], XmNmarginWidth, 2); ac++;
    sprintf(name, "xmbdfed_editor_%ld_form", which);
    rc = XtCreateManagedWidget(name, xmFormWidgetClass, mw, av, ac);

    /*
     * The Font and Glyph info widgets.
     */
    ac = 0;
    XtSetArg(av[ac], XmNtopAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNleftAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNrightAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNshadowType, XmSHADOW_OUT); ac++;
    sprintf(name, "xmbdfed_editor_%ld_info_frame", which);
    frame = XtCreateManagedWidget(name, xmFrameWidgetClass, rc, av, ac);

    sprintf(name, "xmbdfed_editor_%ld_info_form", which);
    form = XtCreateManagedWidget(name, xmFormWidgetClass, frame, 0, 0);

    XtSetArg(av[0], XmNalignment, XmALIGNMENT_END);
    label1 = XtCreateManagedWidget("Font:", xmLabelWidgetClass, form, av, 1);

    ac = 0;
    XtSetArg(av[ac], XmNcolumns, 65); ac++;
    XtSetArg(av[ac], XmNleftAttachment, XmATTACH_WIDGET); ac++;
    XtSetArg(av[ac], XmNleftWidget, label1); ac++;
    sprintf(name, "xmbdfed_editor_%ld_fontname", which);
    ed->fontname = XtCreateManagedWidget(name, xmTextFieldWidgetClass, form,
                                         av, ac);

    XtSetArg(av[0], XmNheight, &ht);
    XtGetValues(ed->fontname, av, 1);
    XtSetArg(av[0], XmNheight, ht);
    XtSetValues(label1, av, 1);

    ac = 0;
    XtSetArg(av[0], XmNalignment, XmALIGNMENT_END);
    XtSetArg(av[ac], XmNtopAttachment, XmATTACH_WIDGET); ac++;
    XtSetArg(av[ac], XmNtopWidget, label1); ac++;
    label2 = XtCreateManagedWidget("Glyph:", xmLabelWidgetClass, form, av, ac);

    /*
     * Adjust the width of the "Font:" widget to match the "Glyph:" widget.
     */
    XtSetArg(av[0], XmNwidth, &wd);
    XtGetValues(label2, av, 1);
    XtSetArg(av[0], XmNwidth, wd);
    XtSetValues(label1, av, 1);

    sprintf(name, "xmbdfed_editor_%ld_glyphinfo", which);
    s = XmStringCreateLtoR("None\nNone", XmSTRING_DEFAULT_CHARSET);

    ac = 0;
    XtSetArg(av[ac], XmNlabelString, s); ac++;
    XtSetArg(av[ac], XmNleftAttachment, XmATTACH_WIDGET); ac++;
    XtSetArg(av[ac], XmNleftWidget, label2); ac++;
    XtSetArg(av[ac], XmNtopAttachment, XmATTACH_WIDGET); ac++;
    XtSetArg(av[ac], XmNtopWidget, ed->fontname); ac++;
    XtSetArg(av[ac], XmNalignment, XmALIGNMENT_BEGINNING); ac++;
    ed->glyphinfo = XtCreateManagedWidget(name, xmLabelWidgetClass, form,
                                          av, ac);
    XmStringFree(s);

    /*
     * The navigation buttons.
     */
    ac = 0;
    XtSetArg(av[ac], XmNtopAttachment, XmATTACH_WIDGET); ac++;
    XtSetArg(av[ac], XmNtopWidget, frame); ac++;
    XtSetArg(av[ac], XmNleftAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNrightAttachment, XmATTACH_FORM); ac++;
    sprintf(name, "xmbdfed_editor_%ld_button_frame", which);
    frame = XtCreateManagedWidget(name, xmFrameWidgetClass, rc, av, ac);

    sprintf(name, "xmbdfed_editor_%ld_button_rowcol", which);
    XtSetArg(av[0], XmNorientation, XmHORIZONTAL);
    rc1 = XtCreateManagedWidget(name, xmRowColumnWidgetClass, frame, av, 1);

    ed->firstpage = XtCreateManagedWidget("First Page",
                                          xmPushButtonWidgetClass, rc1, 0, 0);
    XtAddCallback(ed->firstpage, XmNactivateCallback, GotoFirstPage,
                  (XtPointer) ed->id);

    ed->prevpage = XtCreateManagedWidget("Previous Page",
                                         xmPushButtonWidgetClass, rc1, 0, 0);
    XtAddCallback(ed->prevpage, XmNactivateCallback, GotoPrevPage,
                  (XtPointer) ed->id);

    ed->nextpage = XtCreateManagedWidget("Next Page",
                                         xmPushButtonWidgetClass, rc1, 0, 0);
    XtAddCallback(ed->nextpage, XmNactivateCallback, GotoNextPage,
                  (XtPointer) ed->id);

    ed->lastpage = XtCreateManagedWidget("Last Page",
                                         xmPushButtonWidgetClass, rc1, 0, 0);
    XtAddCallback(ed->lastpage, XmNactivateCallback, GotoLastPage,
                  (XtPointer) ed->id);

    label1 = XtCreateManagedWidget("Page:", xmLabelWidgetClass, rc1, 0, 0);

    ac = 0;
    XtSetArg(av[ac], XmNcolumns, 6); ac++;
    sprintf(name, "xmbdfed_editor_%ld_pageno", which);
    ed->pageno = XtCreateManagedWidget(name, xmTextFieldWidgetClass, rc1,
                                       av, ac);
    XtAddCallback(ed->pageno, XmNactivateCallback, GotoPage,
                  (XtPointer) ed->id);

    label1 = XtCreateManagedWidget("Code:", xmLabelWidgetClass, rc1, 0, 0);

    ac = 0;
    XtSetArg(av[ac], XmNcolumns, 6); ac++;
    sprintf(name, "xmbdfed_editor_%ld_codeno", which);
    ed->code = XtCreateManagedWidget(name, xmTextFieldWidgetClass, rc1,
                                     av, ac);
    XtAddCallback(ed->code, XmNactivateCallback, GotoCode,
                  (XtPointer) ed->id);

    /*
     * Now create the font edit grid.
     */
    ac = 0;
    XtSetArg(av[ac], XmNtopAttachment, XmATTACH_WIDGET); ac++;
    XtSetArg(av[ac], XmNtopWidget, frame); ac++;
    XtSetArg(av[ac], XmNleftAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNrightAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNbottomAttachment, XmATTACH_FORM); ac++;
    if (font != 0) {
        XtSetArg(av[ac], XmNbdfFont, font); ac++;
    }
    XtSetArg(av[ac], XmNpointSize,
             xmbdfed_opts.font_opts.point_size); ac++;
    XtSetArg(av[ac], XmNhorizontalResolution,
             xmbdfed_opts.font_opts.resolution_x); ac++;
    XtSetArg(av[ac], XmNverticalResolution,
             xmbdfed_opts.font_opts.resolution_y); ac++;
    XtSetArg(av[ac], XmNfontSpacing,
             xmbdfed_opts.font_opts.font_spacing); ac++;
    XtSetArg(av[ac], XmNskipBlankPages, xmbdfed_opts.no_blanks); ac++;
    XtSetArg(av[ac], XmNgridOverwriteMode, xmbdfed_opts.overwrite_mode); ac++;
    sprintf(name, "xmbdfed_editor_%ld_fonteditgrid", which);
    ed->fgrid = XtCreateManagedWidget(name, xmuttFontGridWidgetClass, rc,
                                      av, ac);
    XtAddCallback(ed->fgrid, XmNpageCallback, PageChangeCallback,
                  (XtPointer) ed->id);
    XtAddCallback(ed->fgrid, XmNstartSelectionCallback, UpdateGlyphInfo,
                  (XtPointer) ed->id);
    XtAddCallback(ed->fgrid, XmNextendSelectionCallback, UpdateGlyphInfo,
                  (XtPointer) ed->id);
    XtAddCallback(ed->fgrid, XmNactivateCallback, DoGlyphEdit,
                  (XtPointer) ed->id);
    XtAddCallback(ed->fgrid, XmNmodifiedCallback, UpdateModified,
                  (XtPointer) ed->id);

    /*
     * Set the font name text field and add the name change callback only
     * after the font name has been set for the first time.
     */
    XmTextFieldSetString(ed->fontname, XmuttFontGridFontName(ed->fgrid));

    XtAddCallback(ed->fontname, XmNvalueChangedCallback, DoFontNameChange,
                  (XtPointer) ed->id);

    /*
     * Enable or disable the navigation buttons based on the page info
     * of the font.
     */
    XmuttFontGridPageInfo(ed->fgrid, &pi);

    if (pi.previous_page < 0) {
        XtSetSensitive(ed->prevpage, False);
        XtSetSensitive(ed->firstpage, False);
    }
    if (pi.next_page < 0) {
        XtSetSensitive(ed->nextpage, False);
        XtSetSensitive(ed->lastpage, False);
    }

    /*
     * Set the current page number in the text field.
     */
    sprintf(name, "%ld", pi.current_page);
    XmTextFieldSetString(ed->pageno, name);
    XmTextFieldSetCursorPosition(ed->pageno, strlen(name));

    if (XmuttFontGridFontMessages(ed->fgrid) == 0)
      XtSetSensitive(ed->viewmenu.acmsgs, False);
    else
      XtSetSensitive(ed->viewmenu.acmsgs, True);

    /*
     * Make sure the correct code base is set on the code base menu.
     */
    switch (XmuttFontGridCodeBase(ed->fgrid)) {
      case 8: XmToggleButtonSetState(ed->viewmenu.oct, True, False); break;
      case 10: XmToggleButtonSetState(ed->viewmenu.dec, True, False); break;
      case 16: XmToggleButtonSetState(ed->viewmenu.hex, True, False); break;
    }

    /*
     * Enable the Save and Save As menu options if the font has been
     * modified.  Also update the title if necessary.
     */
    if (XmuttFontGridModified(ed->fgrid) == True) {
        sprintf(title, "%s - %s [modified]", app_name, file);
        XtSetArg(av[0], XmNtitle, title);
        XtSetValues(ed->shell, av, 1);
        XtSetSensitive(ed->filemenu.save, True);
    }

    return which;
}

static Boolean
#ifndef _NO_PROTO
_CvtStringToFontSpacing(Display *d, XrmValuePtr args, Cardinal *num_args,
                        XrmValuePtr from, XrmValuePtr to, XtPointer *conv_data)
#else
_CvtStringToFontSpacing(d, args, num_args, from, to, conv_data)
Display *d;
XrmValuePtr args;
Cardinal *num_args;
XrmValuePtr from, to;
XtPointer *conv_data;
#endif
{
    static int fspacing;
    Boolean res;
    char *s;

    res = True;
    if (*num_args != 0)
      XtAppWarningMsg(XtDisplayToApplicationContext(d),
                      "CvtStringToFontSpacing", "wrongParameters",
                      "XmBDFEditorWarning",
                      "String to font spacing needs no extra arguments",
                      0, 0);

    s = (char *) from->addr;
    switch (*s) {
      case 'P': case 'p': fspacing = BDF_PROPORTIONAL; break;
      case 'M': case 'm': fspacing = BDF_MONOSPACE; break;
      case 'C': case 'c': fspacing = BDF_CHARCELL; break;
      default:
          XtDisplayStringConversionWarning(d, (char *) from->addr,
                                           "FontSpacing");
          res = False;
          break;
    }
    if (res == True) {
        if (to->addr == 0)
          to->addr = (XtPointer) &fspacing;
        else if (to->size < sizeof(int))
          res = False;
        else
          *(int *) to->addr = fspacing;
        to->size = sizeof(int);
    }
    return res;
}

static int
#ifndef _NO_PROTO
_XmBDFEditorHandleUnknownOptions(bdf_options_t *opts, char **params,
                                 unsigned long nparams, void *client_data)
#else
_XmBDFEditorHandleUnknownOptions(opts, params, nparams, client_data)
bdf_options_t *opts;
char **params;
unsigned long nparams;
void *client_data;
#endif
{
    MXFEditorOptions *op;

    op = (MXFEditorOptions *) client_data;

    if (nparams == 0)
      return 0;

    if (params[0][0] == 'r' &&
        memcmp(params[0], "resolution", 10) == 0) {
        op->resolution = _bdf_atoul(params[1], 0, 10);
        return 1;
    }

    if (params[0][0] == 's' &&
        memcmp(params[0], "skip_blank_pages", 16) == 0) {
        switch (params[1][0]) {
          case '0': case 'F': case 'f': case 'N': case 'n':
            op->no_blanks = 0;
            break;
          case '1': case 'T': case 't': case 'Y': case 'y':
            op->no_blanks = 1;
            break;
        }
        return 1;
    }

    if (params[0][0] == 'r' &&
        memcmp(params[0], "really_exit", 11) == 0) {
        switch (params[1][0]) {
          case '0': case 'F': case 'f': case 'N': case 'n':
            op->really_exit = 0;
            break;
          case '1': case 'T': case 't': case 'Y': case 'y':
            op->really_exit = 1;
            break;
        }
        return 1;
    }

    if (params[0][0] == 'g' &&
        memcmp(params[0], "grid_overwrite_mode", 19) == 0) {
        switch (params[1][0]) {
          case '0': case 'F': case 'f': case 'N': case 'n':
            op->overwrite_mode = 0;
            break;
          case '1': case 'T': case 't': case 'Y': case 'y':
            op->overwrite_mode = 1;
            break;
        }
        return 1;
    }

    return 0;
}

/*
 * This function does some intialization before any of the X11 stuff happens.
 */
static void
#ifndef _NO_PROTO
_XmBDFEditorSetup(void)
#else
_XmBDFEditorSetup()
#endif
{
    FILE *in;
    char *home, path[BUFSIZ];

    /*
     * Initialize the BDF support library.
     */
    bdf_setup();

    /*
     * Get the default BDF options.
     */
    bdf_default_options(&xmbdfed_opts.font_opts);

    /*
     * Set the default editor options.
     */
    xmbdfed_opts.resolution = 0;
    xmbdfed_opts.no_blanks = xmbdfed_opts.really_exit = 
        xmbdfed_opts.overwrite_mode = 1;

    /*
     * Attempt to load the user config file.
     */
    if ((home = getenv("HOME")) != 0) {
        sprintf(path, "%s/.xmbdfedrc", home);
        if ((in = fopen(path, "r")) != 0) {
            bdf_load_options(in, &xmbdfed_opts.font_opts,
                             _XmBDFEditorHandleUnknownOptions,
                             (void *) &xmbdfed_opts);
            fclose(in);
        }
    }

    /*
     * Update the command line resources from the configuration just loaded.
     * This will cause the current options to be preserved when the command
     * line is parsed by Xt.  The last two command line resources are handled
     * in the normal fashion.
     */
    cmdres[0].default_addr =
        (XtPointer) xmbdfed_opts.font_opts.correct_metrics;
    cmdres[1].default_addr =
        (XtPointer) xmbdfed_opts.font_opts.keep_unencoded;
    cmdres[2].default_addr =
        (XtPointer) xmbdfed_opts.font_opts.keep_comments;

    cmdres[3].default_addr = (XtPointer) xmbdfed_opts.font_opts.pad_cells;
    cmdres[4].default_addr = (XtPointer) xmbdfed_opts.font_opts.point_size;
    cmdres[5].default_addr = (XtPointer) xmbdfed_opts.font_opts.resolution_x;
    cmdres[6].default_addr = (XtPointer) xmbdfed_opts.font_opts.resolution_y;
    cmdres[7].default_addr = (XtPointer) xmbdfed_opts.font_opts.font_spacing;
    cmdres[8].default_addr = (XtPointer) xmbdfed_opts.resolution;
    cmdres[9].default_addr = (XtPointer) xmbdfed_opts.no_blanks;
    cmdres[10].default_addr = (XtPointer) xmbdfed_opts.really_exit;
    cmdres[11].default_addr = (XtPointer) xmbdfed_opts.overwrite_mode;
}

void
#ifndef _NO_PROTO
main(int argc, char *argv[])
#else
main(argc, argv)
int argc;
char *argv[];
#endif
{
    Display *d;
    int i;
    Arg av[1];

    /*
     * Load the config file if it exists and prep the command line resource
     * values with those loaded from the config file.
     */
    _XmBDFEditorSetup();

    /*
     * Create the top level.
     */
    XtSetArg(av[0], XmNallowShellResize, True);
    top = XtAppInitialize(&app, "XmBDFEdit", cmdopts, XtNumber(cmdopts),
                          &argc, argv, fallback_resources, av, 1);

    /*
     * Add the type converter for font spacing values so they can be handled
     * from the command line.
     */
    XtSetTypeConverter(XmRString, "FontSpacing", _CvtStringToFontSpacing,
                       0, 0, XtCacheNone, 0);

    /*
     * Get the recognizable command line options and set the actual font
     * options accordingly.
     */
    XtGetApplicationResources(top, (XtPointer) &xmbdfed_opts, cmdres,
                              XtNumber(cmdres), 0, 0);

    /*
     * Set the horizontal and vertical resolutions if necessary.
     */
    d = XtDisplay(top);
    if (xmbdfed_opts.resolution != 0)
      xmbdfed_opts.font_opts.resolution_x =
          xmbdfed_opts.font_opts.resolution_y = xmbdfed_opts.resolution;

    if (xmbdfed_opts.font_opts.resolution_x == 0)
      xmbdfed_opts.font_opts.resolution_x =
          (long) (((((double) DisplayWidth(d, DefaultScreen(d))) * 25.4) /
                   ((double) DisplayWidthMM(d, DefaultScreen(d)))) + 0.5);

    if (xmbdfed_opts.font_opts.resolution_y == 0)
      xmbdfed_opts.font_opts.resolution_y =
          (long) (((((double) DisplayHeight(d, DefaultScreen(d))) * 25.4) /
                   ((double) DisplayHeightMM(d, DefaultScreen(d)))) + 0.5);

    XtGetApplicationNameAndClass(XtDisplay(top), &app_name, &app_class);

    argc--;
    argv++;

    if (argc == 0)
      (void) MakeEditor(top, 0, True);
    else {
        for (i = 0; i < argc; i++)
          (void) MakeEditor(top, argv[i], True);
    }

    XtRealizeWidget(top);

    /*
     * Make sure the initial font grid has the focus.
     */
    (void) XmProcessTraversal(editors[0].fgrid, XmTRAVERSE_CURRENT);

    XtAppMainLoop(app);
}
